#!/bin/env python

import numpy as np
import scipy.integrate as integrate
import scipy.interpolate as interpolate

#
# Wendland C2 kernel as used in FLAMINGO
#
# Taken from Swift's kernel_hydro.h.
#
kernel_degree=5
kernel_ivals=1
kernel_gamma=1.936492
kernel_constant=21. * np.pi / 2.
kernel_coeffs = np.asarray([4., -15., 20., -10., 0.,  1.,
                            0.,   0.,  0.,   0., 0.,  0.], dtype=float)
kernel_gamma_dim=(kernel_gamma * kernel_gamma * kernel_gamma)
kernel_gamma_inv_dim=(1. / (kernel_gamma * kernel_gamma * kernel_gamma))
kernel_gamma_inv=1./kernel_gamma
kernel_ivals_f=float(kernel_ivals)
kernel_root=kernel_coeffs[kernel_degree] * kernel_constant * kernel_gamma_inv_dim
kernel_norm=(4./3*np.pi) * kernel_gamma_dim


def kernel_eval(u):
    """
    Numpy vectorized kernel function
    """
    u = np.asarray(u, dtype=float)

    # Go to the range [0,1[ from [0,H[
    x = u * kernel_gamma_inv
    
    # Pick the correct branch of the kernel
    temp = (x * kernel_ivals_f).astype(int)
    ind = np.where(temp > kernel_ivals, kernel_ivals, temp)

    coeffs_index = ind * (kernel_degree + 1)

    # First two terms of the polynomial ...
    w = kernel_coeffs[coeffs_index] * x + kernel_coeffs[coeffs_index+1]

    # ... and the rest of them
    for k in range(2, kernel_degree+1):
        w = x * w + kernel_coeffs[coeffs_index+k]

    w = np.asarray(w)
    w[w < 0.0] = 0.0

    return w * kernel_constant * kernel_gamma_inv_dim


def projected_kernel_integrate(u):
    
    u = float(u) # not implemented for array input

    # Avoid integration errors if u is out of range
    if u > kernel_gamma:
        return 0.0

    projected_kernel_integrand = lambda qz : kernel_eval(np.sqrt(u**2+qz**2))

    qz_max = np.sqrt(kernel_gamma**2-u**2)
    qz_min = -qz_max

    result, abserror = integrate.quad(projected_kernel_integrand, qz_min, qz_max)
    return result


class ProjectedKernel:
    """
    Evaluate the projected kernel by interpolating a tabulated function
    """
    def __init__(self):
        """Tabulate the projected kernel"""
        npoints = 1000
        x = np.linspace(0.0, kernel_gamma, npoints)
        y = np.zeros_like(x)
        for i in range(npoints):
            y[i] = projected_kernel_integrate(x[i])
        self.interp = interpolate.interp1d(x, y)

    def __call__(self, u):
        """Evaluate the projected kernel at the input radii"""
        u = np.asarray(u)
        if np.any(u < 0.0):
            raise ValueError("Can't evaluate projected kernel at negative radius!")
        in_range = u < kernel_gamma
        result = np.zeros_like(u, dtype=float)
        result[in_range] = self.interp(u[in_range])
        return result



#!/bin/env python

import sys

import lightcone_io.healpix_maps as hm

if __name__ == "__main__":

    usage="""
    python3 ./lightcone_io_compare_maps.py indir1 indir2 lightcone_nr shell_nr

    indir1      : directory containing the first set of maps
    indir2      : directory containing the second set of maps
    lightcone_nr: which lightcone to compare
    shell_nr    : which shell to compare
"""

    if len(sys.argv) != 5:
        print(usage)
        sys.exit(0)

    indir1       = sys.argv[1]
    indir2       = sys.argv[2]
    lightcone_nr = int(sys.argv[3])
    shell_nr     = int(sys.argv[4])    
    basename = "lightcone%d" % lightcone_nr

    # Find full list of datasets
    fname1 = "%s/%s_shells/shell_%d/%s.shell_%d.0.hdf5" % (indir1, basename, shell_nr, basename, shell_nr)
    map_names1 = hm.get_map_names(fname1)
    fname2 = "%s/%s_shells/shell_%d/%s.shell_%d.0.hdf5" % (indir2, basename, shell_nr, basename, shell_nr)
    map_names2 = hm.get_map_names(fname2)
    datasets = set(map_names1 + map_names2)

    for dataset in datasets:
        maps_match = hm.compare_healpix_maps(indir1, indir2, basename, shell_nr, dataset)
        if maps_match != True:
            raise Exception("Map %s for %s shell %d does not match!" % (dataset, basename, shell_nr))
        else:
            print("Map %s for %s shell %d is ok" % (dataset, basename, shell_nr))



#!/bin/env python

import h5py
import numpy as np
import pytest

from lightcone_io.utils import read_slices, IndexedDatasetReader


# Some simple test cases
test_cases = [
    (),               # Empty index array
    (0,),             # Just one index at the start
    (5000,),          # Just one index somewhere else
    np.arange(10000, dtype=int),       # All elements
    np.arange(1000, dtype=int) + 4000, # A contiguous chunk
    np.arange(1000, dtype=int) * 2,    # Even elements
    np.arange(1000, dtype=int) * 2 + 1,# Odd elements
]

def random_elements(rng, nmax):
    """
    Make a random array of elements to read
    """
    # Decide how many elements to read
    n = rng.integers(0, nmax, 1)[0]
    # Pick the indexes to read
    return rng.integers(0, nmax, size=n)

# Add some random test cases with non-unique, non-sorted indexes
rng = np.random.default_rng(seed=0)
for _ in range(10):
    test_cases.append(random_elements(rng, 10000))


@pytest.mark.parametrize("index", test_cases)
def test_read_slice(dataset, index):
    """
    Check that IndexedDatasetReader reads datasets correctly
    """

    # Read the data using the IndexedDatasetReader
    data1 = IndexedDatasetReader(index).read(dataset)

    # Read all of the data with h5py
    if len(index) > 0:
        # Read everything and have numpy extract the elements we want
        data2 = dataset[...][index,...]
    else:
        # Handle the case of reading zero elements
        shape = (0,)+dataset.shape[1:]
        data2 = np.zeros(shape, dtype=dataset.dtype)

    # Check the result
    assert data1.shape == data2.shape
    assert data1.dtype == data2.dtype
    assert np.all(data1==data2)



#!/bin/env python

import h5py
import numpy as np
import unyt
import virgo.mpi.parallel_hdf5 as phdf5
import virgo.formats.swift as swift
import virgo.util.partial_formatter

from mpi4py import MPI
comm = MPI.COMM_WORLD
comm_rank = comm.Get_rank()
comm_size = comm.Get_size()

# Descriptions for fields generated by the halo lightcone code
extra_descriptions = {
    "Lightcone/HaloCentre" : "Coordinates of the halo at the time of lightcone crossing",
    "Lightcone/Redshift"   : "Redshift of the halo at the time of lightcone crossing",
    "Lightcone/SnapshotNumber" : "Snapshot number from which this halo was taken",
    "Lightcone/TracerPosition" : "Coordinates of the chosen tracer particle at lightcone crossing",
    "InputHalos/SOAPIndex" : "Index of the halo in the input SOAP catalogue",
}

class SOAPCatalogue:
    """
    Class to read SOAP halo catalogues for the halo lightcone
    """
    def __init__(self, halo_format, first_snap, last_snap):
        """
        Determine redshifts of all snapshots
        """
        self.redshift = {}
        if comm_rank == 0:
            for snap_nr in range(first_snap, last_snap+1):
                with h5py.File(halo_format.format(snap_nr=snap_nr), "r") as infile:
                    self.redshift[snap_nr] = float(infile["Cosmology"].attrs["Redshift"][0])
        self.redshift = comm.bcast(self.redshift)
        self.first_snap = first_snap
        self.last_snap = last_snap
        self.halo_format = halo_format
        self.description = {}
        self.description.update(extra_descriptions)

    def read(self, snap_nr, to_read):
        """
        Read in SOAP halos for the specified snapshot.

        Returns a dict of unyt arrays with units constructed from attributes.
        """

        # Create unit registry for this snapshot
        if comm_rank == 0:
            with h5py.File(self.halo_format.format(snap_nr=snap_nr), "r") as infile:
                reg = swift.soap_unit_registry_from_snapshot(infile)
        else:
            reg = None
        reg = comm.bcast(reg)

        # Read the catalogue
        filename = self.halo_format.format(snap_nr=snap_nr)
        mf = phdf5.MultiFile(filename, file_idx=(0,), comm=comm)
        data = mf.read(to_read, read_attributes=True)

        # Store descriptions
        for name in data:
            if "Description" in data[name].attrs:
                self.description[name] = data[name].attrs["Description"]

        # Convert to unyt arrays
        for name in data:
            units = swift.soap_units_from_attributes(data[name].attrs, reg)
            data[name] = unyt.unyt_array(data[name], units=units, registry=reg)

        # Add an array with the index in the SOAP catalogue for each object.
        # This is not the same as the index in HBT because SOAP ignores orphans
        # and reorders the remaining halos.
        nr_halos = data[to_read[0]].shape[0]
        nr_prev_halos = comm.scan(nr_halos) - nr_halos
        soap_index = np.arange(nr_halos, dtype=int) + nr_prev_halos
        soap_index = unyt.unyt_array(soap_index, units="dimensionless", registry=reg)
        data["InputHalos/SOAPIndex"] = soap_index

        return data


class HBTplusCatalogue:
    """
    Class to read HBTplus halo catalogues for the halo lightcone

    This reads HBTplus halo properties and converts them to unyt arrays as if
    we read them from SOAP. This allows construction of lightcone halo
    catalogues without running SOAP on all (or any!) snapshots.
    """
    def __init__(self, halo_format, swift_format, first_snap, last_snap):
        """
        Determine redshifts of all snapshots and system of units
        """
        self.redshift = {}
        self.units = {}
        if comm_rank == 0:
            for snap_nr in range(first_snap, last_snap+1):
                with h5py.File(halo_format.format(snap_nr=snap_nr, file_nr=0), "r") as infile:
                    a = float(infile["Cosmology"]["ScaleFactor"][0])
                    self.redshift[snap_nr] = 1.0/a-1.0
                    if snap_nr == first_snap:
                        self.units["LengthInMpch"] = float(infile["Units"]["LengthInMpch"][0])
                        self.units["MassInMsunh"] = float(infile["Units"]["MassInMsunh"][0])
                        self.units["VelInKmS"] = float(infile["Units"]["VelInKmS"][0])
        self.redshift = comm.bcast(self.redshift)
        self.units = comm.bcast(self.units)
        self.first_snap = first_snap
        self.last_snap = last_snap
        self.halo_format = halo_format
        self.swift_format = swift_format
        self.description = {}
        self.description.update(extra_descriptions)

    def read(self, snap_nr, to_read):
        """
        Read in HBTplus halos for the specified snapshot.

        Returns a dict of unyt arrays with units constructed from attributes.
        Only supports a small subset of SOAP field names.
        """

        # Get the SWIFT unit registry for this snapshot
        if comm_rank == 0:
            filename = self.swift_format.format(snap_nr=snap_nr, file_nr=0)
            with h5py.File(filename, "r") as infile:
                registry = virgo.formats.swift.soap_unit_registry_from_snapshot(infile)
        else:
            registry = None
        registry = comm.bcast(registry)

        # Make a format string for the sub-files in this snapshot
        pf = virgo.util.partial_formatter.PartialFormatter()
        filenames = pf.format(self.halo_format, snap_nr=snap_nr, file_nr=None)

        # Read the full HBTplus catalogue
        mf = phdf5.MultiFile(filenames, file_nr_dataset="NumberOfFiles", comm=comm)
        subhalos = mf.read("Subhalos")

        # Compute halo indexes
        index_offset = comm.scan(len(subhalos)) - len(subhalos)
        index = np.arange(len(subhalos), dtype=int) + index_offset

        # Discard unresolved halos
        keep  = (subhalos["Nbound"] > 1)
        subhalos = subhalos[keep]
        index = index[keep]

        # Construct unyt representation of HBT's units
        h = unyt.Unit("h", registry=registry)
        swift_cmpc = unyt.Unit("a*swift_mpc", registry=registry)
        swift_msun = unyt.Unit("swift_msun", registry=registry)
        hbt_length_unit = (self.units["LengthInMpch"] * swift_cmpc/h).to(swift_cmpc)
        hbt_mass_unit = (self.units["MassInMsunh"] * swift_msun/h).to(swift_msun)

        # Extract the fields we want as unyt arrays using SOAP naming conventions
        data = {}
        for name in to_read:
            if name == "InputHalos/HaloCatalogueIndex":
                data[name] = unyt.unyt_array(index, units="dimensionless", registry=registry)
                self.description[name] = "Index of the halo in the original halo finder catalogue (first halo has index=0)"
            elif name == "InputHalos/HaloCentre":
                pos = subhalos["ComovingMostBoundPosition"].copy()
                data[name] = unyt.unyt_array(pos, units=hbt_length_unit, registry=registry)
                self.description[name] = "The centre of the halo as given by the halo finder"
            elif name == "InputHalos/IsCentral":
                is_central = np.where(subhalos["Rank"]==0, 1, 0)
                data[name] = unyt.unyt_array(is_central, units="dimensionless", registry=registry)
                self.description[name] = "Indicates if halo is central (1) or satellite (0)"
            elif name == "InputHalos/NumberOfBoundParticles":
                nbound = subhalos["Nbound"].copy()
                data[name] = unyt.unyt_array(nbound, units="dimensionless", registry=registry)
                self.description[name] = "Number of bound particles in the halo"
            elif name == "BoundSubhalo/TotalMass":
                mbound = subhalos["Mbound"].copy()
                data[name] = unyt.unyt_array(mbound, units=hbt_mass_unit, registry=registry)
                self.description[name] = "Mass of bound particles in the halo"
            elif name == "InputHalos/HBTplus/TrackId":
                trackid = subhalos["TrackId"].copy()
                data[name] = unyt.unyt_array(trackid, units="dimensionless", registry=registry)
                self.description[name] = "HBTplus TrackId of this halo"
            else:
                raise KeyError("Unsupported HBT property name!")

        return data


if __name__ == "__main__":

    #halo_format = "/cosma8/data/dp004/jch/FLAMINGO/HBT/L1000N0900/HYDRO_FIDUCIAL/SOAP_uncompressed/HBTplus/halo_properties_{snap_nr:04d}.hdf5"
    halo_format = "/cosma8/data/dp004/flamingo/Runs/L1000N0900/DMO_FIDUCIAL/HBT/{snap_nr:03d}/SubSnap_{snap_nr:03d}.{file_nr}.hdf5"
    swift_format = "/cosma8/data/dp004/flamingo/Runs/L1000N0900/DMO_FIDUCIAL/snapshots/flamingo_{snap_nr:04d}/flamingo_{snap_nr:04d}.{file_nr}.hdf5"
    first_snap = 2
    last_snap = 77
    #cat = SOAPCatalogue(halo_format, first_snap, last_snap)
    cat = HBTplusCatalogue(halo_format, swift_format, first_snap, last_snap)

    if comm_rank == 0:
        for snap_nr in range(first_snap, last_snap+1):
            print(snap_nr, cat.redshift[snap_nr])

    # Quantities we need to read in
    to_read = ("InputHalos/cofp", "InputHalos/index")

    data = cat.read(10, to_read)

    if comm_rank == 0:
        print(data)



__all__ = ["ShellArray", "Shell", "HealpixMap", "ParticleLightcone", "IndexedLightconeParticleType", "HaloLightconeFile"]

# Use setuptools_scm to get version from git tags
from importlib.metadata import version, PackageNotFoundError
try:
    __version__ = version("lightcone_io")
except PackageNotFoundError:
    __version__ = "unknown"

# Classes for reading lightcone particle data
from .particle_reader import IndexedLightconeParticleType
from .particle_reader import ParticleLightcone

# Classes for reading healpix maps
from .healpix_maps import Shell, ShellArray, HealpixMap

# Classes for reading halo lightcones
from .halo_reader import HaloLightconeFile



#!/bin/env python

import contextlib
import h5py
import hdfstream
import numpy as np


class LocalOrRemoteFile:
    """
    Mixin class used to open local or remote files
    """
    def set_directory(self, remote_dir=None):
        self._remote_dir = remote_dir

    def open_direct(self, filename):
        if getattr(self, "_remote_dir", None) is None:
            return h5py.File(filename, "r")
        else:
            return self._remote_dir[filename]

    @contextlib.contextmanager
    def open_file(self, filename):
        with self.open_direct(filename) as f:
            yield f


def validate_slices(starts, counts):
    """
    Sanity check the supplied array of slices

    :param starts: 1D array with starting offset of each slice
    :type  starts: np.ndarray
    :param counts: 1D array with length of each slice
    :type  counts: np.ndarray
    """
    if starts.shape != counts.shape:
        raise RuntimeError("start and count arrays must be the same shape")
    if len(starts.shape) != 1 or len(counts.shape) != 1:
        raise RuntimeError("start and count arrays must be 1D")
    if len(starts) > 1:
        if np.any(starts[1:] < starts[:-1]):
            raise RuntimeError("slices must be in ascending order of start index")
        ends = starts + counts
        if np.any(starts[1:] < ends[:-1]):
            raise RuntimeError("slices must not overlap")
    if np.any(counts < 0):
        raise RuntimeError("slices must have non-negative counts")
    if np.any(starts < 0):
        # We don't support negative indexes
        raise RuntimeError("slices must have non-negative starts")


def merge_slices(starts, counts):
    """
    Given a set of slices where slice i starts at index starts[i] and contains
    counts[i] elements, merge any adjacent slices and return new starts and
    counts arrays.

    :param starts: 1D array with starting offset of each slice
    :type  starts: np.ndarray
    :param counts: 1D array with length of each slice
    :type  counts: np.ndarray

    :return: new (starts, counts) tuple with the merged slices
    :rtype: (numpy.ndarray, numpy.ndarray)
    """

    starts = np.asarray(starts, dtype=int)
    counts = np.asarray(counts, dtype=int)

    # First, eliminate any zero length slices
    keep = counts > 0
    starts = starts[keep]
    ends = starts + counts[keep]

    # Determine number of slices
    nr_slices = len(starts)
    if len(ends) != nr_slices:
        raise ValueError("starts and counts arrays must be the same size!")

    # Determine starts to keep: every starting offset which is NOT
    # equal to the end of the previous slice. Always keep the first.
    keep_start = np.ones(nr_slices, dtype=bool)
    keep_start[1:] = (starts[1:] != ends[:-1])

    # Determine ends to keep: every end offset which is NOT equal
    # to the start of the next slice. Always keep the last one.
    keep_end = np.ones(nr_slices, dtype=bool)
    keep_end[:-1] = (ends[:-1] != starts[1:])

    # Discard unwanted elements
    assert len(starts) == len(ends)
    starts = starts[keep_start]
    counts = ends[keep_end] - starts

    return starts, counts


def read_slices_hdfstream(dataset, starts, counts, result=None):
    """
    Read the specified slices from a hdfstream remote dataset. Datasets can
    only be sliced along the first dimension: we always read all elements in
    the remaining dimensions.

    Slices must be in ascending order of starting index and must not overlap.
    Python/numpy style negative indexes from the end of the dataset are not
    supported.

    :param dataset: hdfstream dataset to read from
    :type  dataset: hdfstream.RemoteDataset
    :param starts: 1D array with starting offset of each slice
    :type  starts: np.ndarray
    :param counts: 1D array with length of each slice
    :type  counts: np.ndarray
    :param result: array to hold the result
    :type  result: np.ndarray, or None

    :return: a numpy array with the data
    :rtype: numpy.ndarray
    """
    # Sanity check the slices
    starts = np.asarray(starts, dtype=int)
    counts = np.asarray(counts, dtype=int)
    validate_slices(starts, counts)

    # Merged any adjacent slices
    starts, counts = merge_slices(starts, counts)

    # Make a list of slices to read
    slices = []
    for s, c in zip(starts, counts):
        slices.append(np.s_[s:s+c,...])

    # Request the slices
    if result is None:
        result = dataset.request_slices(slices)
    else:
        dataset.request_slices(slices, dest=result)
    return result


def read_slices_h5py(dataset, starts, counts, result=None):
    """
    Read the specified slices from a HDF5 dataset. Uses h5py low level calls
    to read the slices with a single H5Dread(). Datasets can only be sliced
    along the first dimension: we always read all elements in the remaining
    dimensions.

    Slices must be in ascending order of starting index and must not overlap.
    Python/numpy style negative indexes from the end of the dataset are not
    supported.

    :param dataset: HDF5 dataset to read from
    :type  dataset: h5py.Dataset
    :param starts: 1D array with starting offset of each slice
    :type  starts: np.ndarray
    :param counts: 1D array with length of each slice
    :type  counts: np.ndarray
    :param result: array to hold the result
    :type  result: np.ndarray, or None

    :return: a numpy array with the data
    :rtype: numpy.ndarray
    """

    # Sanity check the slices
    starts = np.asarray(starts, dtype=int)
    counts = np.asarray(counts, dtype=int)
    validate_slices(starts, counts)

    # Merged any adjacent slices
    starts, counts = merge_slices(starts, counts)

    # Get dataset handle
    dataset_id = dataset.id

    # Get file dataspace handle
    file_space_id = dataset_id.get_space()
    file_shape = file_space_id.get_simple_extent_dims()

    # Select the slices to read
    nr_in_first_dim = 0
    file_space_id.select_none()
    for start, count in zip(starts, counts):
        if count > 0:
            # Select this slice
            slice_start = tuple([start,]+[0 for fs in file_shape[1:]])
            slice_count = tuple([count,]+[fs for fs in file_shape[1:]])
            file_space_id.select_hyperslab(slice_start, slice_count, op=h5py.h5s.SELECT_OR)
            nr_in_first_dim += count

    # Allocate the output array, if necessary
    result_shape = [nr_in_first_dim,]+list(file_shape[1:])
    result_shape = tuple([int(rs) for rs in result_shape])
    if result is None:
        result = np.ndarray(result_shape, dtype=dataset.dtype)

    # Output array must be C contiguous
    if not result.flags['C_CONTIGUOUS']:
        raise RuntimeError("Can only read into C contiguous arrays!")

    # Output array must have the expected number of elements
    nr_selected = file_space_id.get_select_npoints()
    if nr_selected != result.size:
        raise RuntimeError("Output buffer is not the right size for the selected slices!")

    # The output array must have the expected shape (could be wrong if it was passed in)
    if result.shape != result_shape:
        raise RuntimeError("Output buffer has the wrong shape!")

    # If we selected any elements, read the data
    if nr_in_first_dim > 0:
        mem_space_id = h5py.h5s.create_simple(result_shape)
        dataset_id.read(mem_space_id, file_space_id, result)

    return result


def read_slices(dataset, starts, counts, result=None):
    """
    Read the specified slices from a h5py dataset or a hdfstream remote
    dataset.

    :param dataset: dataset to read from
    :type  dataset: h5py.Dataset or hdfstream.RemoteDataset
    :param starts: 1D array with starting offset of each slice
    :type  starts: np.ndarray
    :param counts: 1D array with length of each slice
    :type  counts: np.ndarray
    :param result: array to hold the result
    :type  result: np.ndarray, or None

    :return: a numpy array with the data
    :rtype: numpy.ndarray
    """
    if isinstance(dataset, h5py.Dataset):
        return read_slices_h5py(dataset, starts, counts, result)
    elif isinstance(dataset, hdfstream.RemoteDataset):
        return read_slices_hdfstream(dataset, starts, counts, result)
    else:
        raise RuntimeError("Unsupported dataset type")


class IndexedDatasetReader:

    def __init__(self, index, sorted_and_unique=False):
        """
        Class for reading specified indexes from HDF5 datasets. Here we assume
        that the requested indexes are likely to include runs of consecutive
        values and so can be efficiently handled using hyperslab reads. The
        array of indexes is converted into (start, count) pairs and datasets
        are read using read_slices().

        The supplied indexes are in the first dimension. We read all data in any
        subsequent dimensions. Indexes must be unique and in ascending order if
        sorted_and_unique is True.

        An instance of this class can be used to read the same elements from
        multiple datasets.

        :param index: 1D array with indexes to read in the first dimension
        :type  index: np.ndarray
        :param sorted_and_unique: set to True if index values are sorted and unique
        :type  sorted_and_unique: bool
        """
        # Get sorted, unique indexes if necessary
        index = np.asarray(index, dtype=int)
        if sorted_and_unique:
            self.unique_index = index
            self.inverse_index = None
        else:
            self.unique_index, self.inverse_index = np.unique(index, return_inverse=True)

        # Every index is a range of length one. Merge any adjacent ranges.
        self.starts, self.counts = merge_slices(self.unique_index, np.ones(len(self.unique_index), dtype=int))

    def read(self, dataset):
        """
        Read the specified indexes from a HDF5 dataset.

        :param dataset: HDF5 dataset to read from
        :type  dataset: h5py.Dataset

        :return: a numpy array with the data read from the dataset
        :rtype: numpy.ndarray
        """
        # Read in the specified ranges
        result = read_slices(dataset, self.starts, self.counts)

        # And put the result into the order in which the indexes were requested
        if self.inverse_index is not None:
            result = result[self.inverse_index,...]
        return result


class SlicedDatasetReader:

    def __init__(self, starts, counts):
        """
        Class for reading specified slices from HDF5 datasets. Datasets are
        read using read_slices(). The supplied slices are in the first
        dimension. We read all data in any subsequent dimensions.

        An instance of this class can be used to read the same elements from
        multiple datasets.

        :param starts: 1D array with starting offset of each slice
        :type  starts: np.ndarray
        :param counts: 1D array with length of each slice
        :type  counts: np.ndarray
        """
        # Merge and store any adjacent ranges
        self.starts, self.counts = merge_slices(starts, counts)

    def read(self, dataset):
        """
        Read the specified indexes from a HDF5 dataset.

        :param dataset: HDF5 dataset to read from
        :type  dataset: h5py.Dataset

        :return: a numpy array with the data read from the dataset
        :rtype: numpy.ndarray
        """
        # Read in the specified ranges
        return read_slices(dataset, self.starts, self.counts)

    def count(self):
        """
        Return the number of selected elements

        :return: number of selected elements
        :rtype: int
        """
        return np.sum(self.counts, dtype=int)


def match(arr1, arr2, arr2_sorted=False, arr2_index=None):
    """
    For each element in arr1 return the index of the element with the
    same value in arr2, or -1 if there is no element with the same value.
    Setting arr2_sorted=True will save some time if arr2 is already sorted
    into ascending order.

    A precomputed sorting index for arr2 can be supplied using the
    arr2_index parameter. This can save time if the routine is called
    repeatedly with the same arr2 but arr2 is not already sorted.

    It is assumed that each element in arr1 only occurs once in arr2.
    """

    # Check for the case where we're searching an empty arr2 - can't be any matches
    if len(arr2) == 0:
        return -np.ones(len(arr1), dtype=int)

    # Workaround for a numpy bug (<=1.4): ensure arrays are native endian
    # because searchsorted ignores endian flag
    if not(arr1.dtype.isnative):
        arr1_n = np.asarray(arr1, dtype=arr1.dtype.newbyteorder("="))
    else:
        arr1_n = arr1
    if not(arr2.dtype.isnative):
        arr2_n = np.asarray(arr2, dtype=arr2.dtype.newbyteorder("="))
    else:
        arr2_n = arr2

    # Sort arr2 into ascending order if necessary
    tmp1 = arr1_n
    if arr2_sorted:
        tmp2 = arr2_n
        idx = slice(0,len(arr2_n))
    else:
        if arr2_index is None:
            idx = np.argsort(arr2_n)
            tmp2 = arr2_n[idx]
        else:
            # Use supplied sorting index
            idx = arr2_index
            tmp2 = arr2_n[arr2_index]

    # Find where elements of arr1 are in arr2
    ptr  = np.searchsorted(tmp2, tmp1)

    # Make sure all elements in ptr are valid indexes into tmp2
    # (any out of range entries won't match so they'll get set to -1
    # in the next bit)
    ptr[ptr>=len(tmp2)] = 0
    ptr[ptr<0]          = 0

    # Return -1 where no match is found
    ind  = tmp2[ptr] != tmp1
    ptr[ind] = -1

    # Put ptr back into original order
    ind = np.arange(len(arr2_n))[idx]
    ptr = np.where(ptr>= 0, ind[ptr], -1)

    return ptr



#!/bin/env python

import io
import h5py
import hdfstream
import numpy as np
import pytest


def pytest_addoption(parser):
    parser.addoption("--hdfstream-server", default=None,
                     help="Hdfstream server URL for tests")
    parser.addoption("--hdfstream-dir", default="/",
                     help="Hdfstream directory with test data")


@pytest.fixture(scope='module', params=[False, True])
def remote_dir(request):
    if request.param:
        # Will read local files
        yield None
    else:
        # Will read remote files
        server = request.config.getoption("--hdfstream-server")
        dirname = request.config.getoption("--hdfstream-dir")
        if server is None or dirname is None:
            pytest.skip("Need hdfstream server and directory names for test")
        yield hdfstream.open(server, dirname)


@pytest.fixture(scope='module', params=[(), (0,), (1,), (3,), (3,2)])
def dataset(request):
    """
    Make an in-memory dataset for testing.

    Here we test slicing along the first dimension. The fixture param here
    contains the dimensions of the dataset in the remaining dimensions.
    """
    # Size in the first dimension
    nr_first_dim = 10000

    # Full shape of the test dataset
    shape = (nr_first_dim,)+request.param

    # Compute total number of elements
    nr_total = nr_first_dim
    for n in shape[1:]:
        nr_total *= n
    # Make an array with the test data
    data = np.arange(nr_total, dtype=int).reshape(shape)

    # Write the HDF5 dataset
    bio = io.BytesIO()
    with h5py.File(bio, 'w') as f:
        f["data"] = data

    # Reopen in read only mode
    bio.seek(0)
    with h5py.File(bio, 'r') as f:
        yield f["data"]




#!/bin/env python

import io
import h5py
import numpy as np
import pytest

from lightcone_io.utils import merge_slices, validate_slices


def count_times_selected(n, starts, counts):
    """
    Count how many times each of n array elements is in a slice
    """
    nr_selections = np.zeros(n, dtype=int)
    for s, c in zip(starts, counts):
        nr_selections[s:s+c] += 1

    # Elements should all be selected 0-1 times
    assert np.all(nr_selections>=0)
    assert np.all(nr_selections<=1)

    return nr_selections


def assert_slices_equivalent(slicespec1, slicespec2):
    """
    Check that slicespec1 and slicespec2 specify the same elements

    Each slicespec is a (starts, counts) tuple where starts is an array
    with the offset to each slice and counts is an array with the length
    of each slice.
    """
    starts1, counts1 = slicespec1
    validate_slices(starts1, counts1)
    starts2, counts2 = slicespec2
    validate_slices(starts2, counts2)

    # Determine the maximum array index
    if len(starts1) > 0:
        n1 = np.amax(starts1+counts1)
    else:
        n1 = 0
    if len(starts2) > 0:
        n2 = np.amax(starts2+counts2)
    else:
        n2 = 0
    n = max(n1, n2)

    # Check that the slices select the same elements
    selections1 = count_times_selected(n, starts1, counts1)
    selections2 = count_times_selected(n, starts2, counts2)
    assert np.all(selections1==selections2)


def random_slices(nr_slices, max_length, rng):
    """
    Generate some random test cases for test_merge_slices()
    """
    # Maximum length of a slice or gap between slices
    max_length = 20

    # Draw lengths of slices and gaps between them
    lengths = rng.integers(0, max_length+1, size=2*nr_slices)

    # Generate the slices
    all_starts = []
    all_counts = []
    offset = 0
    for i in range(nr_slices):
        # Skip some elements before this slice
        offset += lengths[2*i+0]
        # Define this slice
        start = offset
        count = lengths[2*i+1]
        # Advance past this slice
        offset += count
        # Store the slice
        all_starts.append(start)
        all_counts.append(count)
    return np.asarray(all_starts, dtype=int), np.asarray(all_counts, dtype=int)


# Some simple test cases for test_merge_slices()
test_slices = [
    ([],        []),        # an empty list of slices
    ([0,],      [0,]),      # one zero length slice
    ([0, 1],    [0, 0]),    # two zero length slices
    ([0,],      [10,]),     # a single slice
    ([5,],      [10,]),     # a single slice with non-zero start
    ([0, 5],    [5, 5]),    # two slices which can be merged
    ([0, 5, 5], [5, 0, 5]), # two slices which can be merged, separated by a zero size slice
    ([0, 6],    [5, 5]),    # two slices which cannot be merged
]

# Add some random test cases
rng = np.random.default_rng(seed=0)
for _ in range(50):
    nr_slices, max_length = rng.integers(0, 10, size=2)
    test_slices.append(random_slices(nr_slices, max_length, rng))


@pytest.mark.parametrize("starts,counts", test_slices)
def test_merge_slices(starts, counts):
    """
    Check that merging slices does not change the selected array indexes
    """
    starts = np.asarray(starts, dtype=int)
    counts = np.asarray(counts, dtype=int)
    merged_starts, merged_counts = merge_slices(starts, counts)
    assert_slices_equivalent((starts, counts), (merged_starts, merged_counts))



#!/bin/env python3

import sys
import os.path
import shutil

import h5py

from mpi4py import MPI
comm = MPI.COMM_WORLD
comm_rank = comm.Get_rank()
comm_size = comm.Get_size()

import lightcone_io.healpix_maps as hm


def combine_maps_mpi(indir, outdir, basenames):
    
    if comm_rank == 0:
        # Read the index files to get number of shells
        shells = []
        for basename in basenames:
            fname = indir+("/%s_index.hdf5" % basename)
            with h5py.File(fname, "r") as infile:
                nr_shells = infile["Lightcone"].attrs["nr_shells"][0]
            print("Lightcone %s has %d shells" % (basename, nr_shells))
            for shell_nr in range(nr_shells):
                shells.append((basename, shell_nr))
    else:
        shells = None
    shells = comm.bcast(shells)

    # Merge the maps
    for index, (basename, shell_nr) in enumerate(shells):
        if index % comm_size == comm_rank:
            hm.combine_healpix_maps(indir, basename, shell_nr, outdir)

    # Copy and update the index files if necessary
    if comm_rank == 0:
        for basename in basenames:
            input_index  = indir+("/%s_index.hdf5" % basename)
            output_index = outdir+("/%s_index.hdf5" % basename)
            if not(os.path.exists(output_index)):
                shutil.copyfile(input_index, output_index)
            with h5py.File(output_index, "r+") as indexfile:
                indexfile["Lightcone"].attrs["nr_files_per_shell"] = [1,]


if __name__ == "__main__":
    
    usage="""
    Usage: python3 -m mpi4py lightcone_io_combine_maps.py \\
              basedir outdir basename1 basename2 ...

    basedir      : location of the input lightcones
    outdir       : where to write the output
    basename*    : basenames of lightcones to process
"""

    args = {}
    if comm_rank == 0:
        if len(sys.argv) < 4:
            print(usage)
            args = None
        else:
            args["indir"] = sys.argv[1]
            args["outdir"] = sys.argv[2]
            args["basenames"] = sys.argv[3:]
    args = comm.bcast(args)

    if args is None:
        MPI.Finalize()
        sys.exit(0)

    combine_maps_mpi(args["indir"], args["outdir"], args["basenames"])

    comm.barrier()
    if comm_rank == 0:
        print("All shells done for lightcones ", args["basenames"])



#!/bin/env python

import os
import sys
import argparse
import time
t0 = time.time()

import numpy as np
import h5py
import scipy.spatial

from mpi4py import MPI
comm = MPI.COMM_WORLD
comm_size = comm.Get_size()
comm_rank = comm.Get_rank()

import virgo.util.match as match
import virgo.mpi.parallel_hdf5 as phdf5
import virgo.mpi.parallel_sort as psort
import virgo.mpi.util as mpi_util


# Constants to identify methods for dealing with particles in multiple halos
FRACTIONAL_RADIUS=0
MOST_MASSIVE=1
LEAST_MASSIVE=2
overlap_methods = {
    "fractional-radius" : FRACTIONAL_RADIUS,
    "most-massive"      : MOST_MASSIVE,
    "least-massive"     : LEAST_MASSIVE,
    }


def message(m):
    if comm_rank == 0:
        t1 = time.time()
        elapsed = t1-t0
        print(f"{elapsed:.1f}s: {m}")


def read_lightcone_halo_positions_and_radii(args, radius_name, mass_name):
    """
    Read in the lightcone halo catalogue and cross reference with SOAP
    to find the SO radius and mass for each halo in the lightcone.

    Assumes that positions in the lightcone are comoving and in the
    same units as SOAP (except for the expansion factor dependence).
    """

    # Parallel read the halo catalogue: need (x,y,z), snapnum, id
    message("Reading lightcone halo catalogue")
    halo_lightcone_datasets = ("LightconeXcminpot", "LightconeYcminpot", "LightconeZcminpot", "SnapNum", "ID")
    mf = phdf5.MultiFile(args.halo_lightcone_filenames, file_nr_attr=("Header", "NumberOfFiles"), comm=comm)
    halo_lightcone_data = mf.read(halo_lightcone_datasets, group="Subhalo", read_attributes=True)

    # Store index in halo lightcone of each halo
    nr_local_halos = len(halo_lightcone_data["ID"])
    offset = comm.scan(nr_local_halos) - nr_local_halos
    halo_lightcone_data["IndexInHaloLightcone"] = np.arange(nr_local_halos, dtype=int) + offset

    # Repartition halos for better load balancing
    message("Repartition halo catalogue")
    nr_local_halos = len(halo_lightcone_data["ID"])
    nr_total_halos = comm.allreduce(nr_local_halos)
    nr_desired = np.zeros(comm_size, dtype=int)
    nr_desired[:] = nr_total_halos // comm_size
    nr_desired[:nr_total_halos % comm_size] += 1
    assert np.sum(nr_desired) == nr_total_halos
    for name in halo_lightcone_data:
        halo_lightcone_data[name] = psort.repartition(halo_lightcone_data[name], nr_desired, comm=comm)

    # Merge x/y/z into a single array
    halo_pos = np.column_stack((halo_lightcone_data["LightconeXcminpot"],
                                halo_lightcone_data["LightconeYcminpot"],
                                halo_lightcone_data["LightconeZcminpot"]))
    del halo_lightcone_data["LightconeXcminpot"]
    del halo_lightcone_data["LightconeYcminpot"]
    del halo_lightcone_data["LightconeZcminpot"]
    halo_lightcone_data["Pos_minpot"] = halo_pos 
    del halo_pos

    # The input catalogue is ordered by redshift, but we want a mix of redshifts on each rank
    message("Reassign halos to MPI ranks")
    nr_local_halos = len(halo_lightcone_data["ID"])
    rng = np.random.default_rng()
    sort_key = rng.integers(comm_size, size=nr_local_halos, dtype=np.int32)
    order = psort.parallel_sort(sort_key, comm=comm, return_index=True)
    for name in sorted(halo_lightcone_data):
        psort.fetch_elements(halo_lightcone_data[name], order, result=halo_lightcone_data[name], comm=comm)

    # Sort locally by snapnum
    message("Sorting local lightcone halos by snapshot")
    order = np.argsort(halo_lightcone_data["SnapNum"])
    for name in halo_lightcone_data:
        halo_lightcone_data[name] = halo_lightcone_data[name][order,...]

    # Find range of local halos at each snapshot
    message("Identifying halos at each snapshot")
    unique_snap, snap_offset, snap_count = np.unique(halo_lightcone_data["SnapNum"],
                                                     return_index=True, return_counts=True)

    # Find full range of snapshots across all MPI ranks
    min_snap = comm.allreduce(np.amin(unique_snap), op=MPI.MIN)
    max_snap = comm.allreduce(np.amax(unique_snap), op=MPI.MAX)

    # Make snapnum, count and offset arrays which include snapshots not present on this rank:
    # We're going to do collective reads of the SOAP outputs so all ranks need to agree on
    # what range of snapshots to do.
    nr_snaps = max_snap - min_snap + 1
    unique_snap_all = np.arange(min_snap, max_snap+1, dtype=int)
    snap_offset_all = np.zeros(nr_snaps, dtype=int)
    snap_count_all  = np.zeros(nr_snaps, dtype=int)
    for us, so, sc in zip(unique_snap, snap_offset, snap_count):
        i = us - min_snap
        assert unique_snap_all[i] == us
        snap_offset_all[i] = so
        snap_count_all[i] = sc

    # Allocate storage for radius of each lightcone halo
    nr_halos = len(halo_lightcone_data["ID"])
    halo_lightcone_data[radius_name] = None # Don't know dtype for radius array yet

    # Loop over snapshots
    for snapnum in unique_snap_all:

        # Datasets to read from SOAP
        soap_datasets = ("VR/ID", radius_name, mass_name)

        # Read the SOAP catalogue for this snapshot
        message(f"Reading SOAP output for snapshot {snapnum}")
        mf = phdf5.MultiFile(args.soap_filenames % {"snap_nr" : snapnum}, file_idx=(0,), comm=comm)
        soap_data = mf.read(soap_datasets, read_attributes=True)

        # Get the expansion factor of this snapshot
        if comm_rank == 0:
            with h5py.File(args.soap_filenames % {"snap_nr" : snapnum}, "r") as infile:
                a = float(infile["SWIFT"]["Header"].attrs["Scale-factor"])
        else:
            a = None
        a = comm.bcast(a)

        # Ensure radii are in comoving units
        radius_a_exponent = float(soap_data[radius_name].attrs["a-scale exponent"])
        soap_data[radius_name] *= a**(radius_a_exponent-1.0)

        # Match lightcone halos at this snapshot to SOAP halos by ID
        message("Finding lightcone halos in SOAP output")
        i1 = snap_offset_all[snapnum-min_snap]
        i2 = snap_offset_all[snapnum-min_snap] + snap_count_all[snapnum-min_snap]
        assert np.all(halo_lightcone_data["SnapNum"][i1:i2] == snapnum)
        ptr = psort.parallel_match(halo_lightcone_data["ID"][i1:i2], soap_data["VR/ID"], comm=comm)
        assert np.all(ptr>=0) # All halos in the lightcone should be found in SOAP

        # Allocate storage for radii now that we know what dtype SOAP uses
        if halo_lightcone_data[radius_name] is None:
            radius_dtype = soap_data[radius_name].dtype
            halo_lightcone_data[radius_name] = phdf5.AttributeArray(-np.ones(nr_halos, dtype=radius_dtype),
                                                                    attrs=soap_data[radius_name].attrs)
            mass_dtype = soap_data[mass_name].dtype
            halo_lightcone_data[mass_name] = phdf5.AttributeArray(-np.ones(nr_halos, dtype=mass_dtype),
                                                                  attrs=soap_data[mass_name].attrs)

        message("Storing SO radii for lightcone halos at this snapshot")
        psort.fetch_elements(soap_data[radius_name], ptr,
                             result=halo_lightcone_data[radius_name][i1:i2], comm=comm)

        message("Storing SO masses for lightcone halos at this snapshot")
        psort.fetch_elements(soap_data[mass_name], ptr,
                             result=halo_lightcone_data[mass_name][i1:i2], comm=comm)

    # All halos should have been assigned a radius
    assert np.all(halo_lightcone_data[radius_name] >= 0)
    assert np.all(halo_lightcone_data[mass_name] >= 0)

    return halo_lightcone_data


def read_lightcone_index(args):
    """
    Read the index file and determine names of all particle files
    and which particle types are present
    """
    
    # Particle types which may be in the lightcone:
    type_names = ("BH", "DM", "Gas", "Neutrino", "Stars")
    type_z_range = {}

    # Now, find the lightcone particle output and read the index info
    index_file = args.lightcone_dir+"/"+args.lightcone_base+"_index.hdf5"
    if comm_rank == 0:
        with h5py.File(index_file, "r") as index:
            lc = index["Lightcone"]
            nr_mpi_ranks = int(lc.attrs["nr_mpi_ranks"])
            final_file_on_rank = lc.attrs["final_particle_file_on_rank"]
            for tn in type_names:
                min_z = float(lc.attrs["minimum_redshift_"+tn])
                max_z = float(lc.attrs["maximum_redshift_"+tn])
                if max_z > min_z:
                    type_z_range[tn] = (min_z, max_z)
    else:
        nr_mpi_ranks = None
        final_file_on_rank = None
        type_z_range = None
    nr_mpi_ranks, final_file_on_rank, type_z_range = comm.bcast((nr_mpi_ranks, final_file_on_rank, type_z_range))

    # Report which particle types we found
    for name in type_z_range:
        min_z, max_z = type_z_range[name]
        message(f"have particles for type {name} from z={min_z} to z={max_z}")

    # Make a full list of files to read
    all_particle_files = []
    for rank_nr in range(nr_mpi_ranks):
        for file_nr in range(final_file_on_rank[rank_nr]+1):
            filename = f"{args.lightcone_dir}/{args.lightcone_base}_particles/{args.lightcone_base}_{file_nr:04d}.{rank_nr}.hdf5"
            all_particle_files.append(filename)

    return type_z_range, all_particle_files


def compute_particle_group_index(halo_id, halo_pos, halo_radius, halo_mass, part_pos,
                                 overlap_method):
    """
    Tag particles which are within the SO radius of a halo
    """

    # Assign indexes to the particles so we can restore their ordering later
    nr_particles = part_pos.shape[0]
    offset = comm.scan(nr_particles) - nr_particles
    part_index = np.arange(nr_particles, dtype=np.int64) + offset

    nr_halos = halo_pos.shape[0]
    nr_halos_total = comm.allreduce(nr_halos)
    nr_particles_total = comm.allreduce(nr_particles)
    message(f"Have {nr_particles_total} particles and {nr_halos_total} halos")

    # Will split the halos and particles by x coordinate, with a roughly
    # constant number of particles per rank. First, sort the particles by x.
    message("Sorting particles by x coordinate")
    sort_key = part_pos[:,0].copy()
    order = psort.parallel_sort(sort_key, return_index=True, comm=comm)
    del sort_key
    psort.fetch_elements(part_pos, order, result=part_pos, comm=comm)
    psort.fetch_elements(part_index, order, result=part_index, comm=comm)
    del order

    # Find the maximum halo radius
    local_max_radius = np.amax(halo_radius)
    max_radius = comm.allreduce(local_max_radius, op=MPI.MAX)

    # Find maximum distance to any particle
    local_max_particle_distance = np.amax(np.sqrt(np.sum(part_pos**2, axis=1)))
    max_particle_distance = comm.allreduce(local_max_particle_distance, op=MPI.MAX)

    # Find the subset of halos which can overlap the particle distribution:
    # This helps in case the halo lightcone goes out to much higher redshift
    # than the particle lightcone.
    halo_distance = np.sqrt(np.sum(halo_pos**2, axis=1))
    within_distance = halo_distance < (max_particle_distance + max_radius)
    halo_pos = halo_pos[within_distance,:]
    halo_id = halo_id[within_distance]
    halo_radius = halo_radius[within_distance]
    halo_mass = halo_mass[within_distance]
    nr_halos_left = comm.allreduce(halo_id.shape[0], op=MPI.SUM)
    message(f"Halos within redshift range = {nr_halos_left} of {nr_halos_total}")

    # Determine the range of x coordinates of halos which could overlap particles on this rank
    local_x_min = np.amin(part_pos[:,0]) - max_radius
    x_min_on_rank = np.asarray(comm.allgather(local_x_min), dtype=part_pos.dtype)
    local_x_max = np.amax(part_pos[:,0]) + max_radius
    x_max_on_rank = np.asarray(comm.allgather(local_x_max), dtype=part_pos.dtype)

    # Sort local halos by x coordinate
    message("Sorting local lightcone halos by x coordinate")
    order = np.argsort(halo_pos[:,0])
    halo_pos = halo_pos[order,:]
    halo_id = halo_id[order]
    halo_radius = halo_radius[order]
    halo_mass = halo_mass[order]
    del order

    # Determine what range of halos needs to be sent to each MPI rank
    first_halo_for_rank = np.searchsorted(halo_pos[:,0], x_min_on_rank, side="left")
    last_halo_for_rank = np.searchsorted(halo_pos[:,0], x_max_on_rank, side="right")
    nr_halos_for_rank = last_halo_for_rank - first_halo_for_rank
    nr_halos_for_rank_total = comm.allreduce(nr_halos_for_rank)
    total_nr_halos_read = comm.allreduce(halo_pos.shape[0])
    total_nr_halos_sent = np.sum(nr_halos_for_rank_total)
    duplication_factor = total_nr_halos_sent / total_nr_halos_read
    message(f"Minimum halos on rank after exchange = {np.amin(nr_halos_for_rank_total)}")
    message(f"Maximum halos on rank after exchange = {np.amax(nr_halos_for_rank_total)}")
    message(f"Duplication factor = {duplication_factor}")

    # Compute lengths and offsets for alltoallv halo exchange
    send_offset = first_halo_for_rank
    send_count = nr_halos_for_rank
    recv_count = np.asarray(comm.alltoall(send_count), dtype=send_count.dtype)
    recv_offset = np.cumsum(recv_count) - recv_count

    # Exchange halo IDs
    message("Exchanging halo IDs")
    halo_id_recv = np.empty_like(halo_id, shape=np.sum(recv_count))
    psort.my_alltoallv(halo_id, send_count, send_offset,
                       halo_id_recv, recv_count, recv_offset,
                       comm=comm)
    halo_id = halo_id_recv
    del halo_id_recv
    comm.barrier()

    # Exchange halo radii
    message("Exchanging halo radii")
    halo_radius_recv = np.empty_like(halo_radius, shape=np.sum(recv_count))
    psort.my_alltoallv(halo_radius, send_count, send_offset,
                       halo_radius_recv, recv_count, recv_offset,
                       comm=comm)
    halo_radius = halo_radius_recv
    del halo_radius_recv
    comm.barrier()

    # Exchange halo masses
    message("Exchanging halo masses")
    halo_mass_recv = np.empty_like(halo_mass, shape=np.sum(recv_count))
    psort.my_alltoallv(halo_mass, send_count, send_offset,
                       halo_mass_recv, recv_count, recv_offset,
                       comm=comm)
    halo_mass = halo_mass_recv
    del halo_mass_recv
    comm.barrier()
    
    # Exchange halo positions:
    # These are vectors so flatten, exchange then restore shape
    message("Exchanging halo positions")
    halo_pos.shape = (-1,)
    halo_pos_recv = np.empty_like(halo_pos, shape=3*np.sum(recv_count))
    psort.my_alltoallv(halo_pos, send_count*3, send_offset*3,
                       halo_pos_recv, recv_count*3, recv_offset*3,
                       comm=comm)
    halo_pos = halo_pos_recv
    halo_pos.shape = (-1, 3)
    del halo_pos_recv
    comm.barrier()

    # Build a kdtree with the local particles
    message("Building kdtree")
    tree = scipy.spatial.KDTree(part_pos)

    # Allocate output array for the particle halo IDs etc
    nr_parts = part_pos.shape[0]
    part_halo_id = -np.ones(nr_parts, dtype=np.int64)       # ID of halo particle is assigned to
    part_halo_mass = np.ndarray(nr_parts, dtype=np.float32) # Mass of the halo
    if overlap_method == LEAST_MASSIVE:
        # Looking for least massive halo, so initialize mass to huge value
        part_halo_mass[:] = np.finfo(part_halo_mass.dtype).max
    else:
        # Looking for most massive halo or not using mass, so initialize mass to -1
        part_halo_mass[:] = -1
    part_halo_r_frac_2 = -np.ndarray(nr_parts, dtype=np.float32) # Smallest ((Particle radius)/(halo r200))**2 so far
    part_halo_r_frac_2[:] = np.finfo(part_halo_r_frac_2.dtype).max  # Initialize min. fractional radius to huge value

    # Report maximum halo radius
    max_radius = comm.allreduce(np.amax(halo_radius), op=MPI.MAX)
    message(f"Maximum halo radius = {max_radius}")

    # Loop over local halos
    nr_assigned = 0
    message("Assigning halo IDs to particles")
    for i in range(len(halo_id)):
            
        # Identify particles within this halo's radius
        idx = np.asarray(tree.query_ball_point(halo_pos[i,:], halo_radius[i]), dtype=int)

        # Compute radius squared for each particle
        r_part_2 = np.sum((part_pos[idx,:] - halo_pos[i,:])**2.0, axis=1)
        
        # Compute ((particle radius)/(halo radius))**2
        r_frac_2 = r_part_2 / (halo_radius[i]**2)

        # Identify particles to update
        if overlap_method == FRACTIONAL_RADIUS:
            # Assign particles to this halo if (particle radius)/(halo radius) is smaller
            # than the smallest value so far
            to_update = (r_frac_2 < part_halo_r_frac_2[idx])
        elif overlap_method == MOST_MASSIVE:
            # Assign particles to this halo if this is the most massive halo the particle
            # has been found to be in so far
            to_update = (halo_mass[i] > part_halo_mass[idx])
        elif overlap_method == LEAST_MASSIVE:
            # Assign particles to this halo if this is the least massive halo the particle
            # has been found to be in so far
            to_update = (halo_mass[i] < part_halo_mass[idx])
        else:
            raise ValueError("Unrecognized value of overlap_method")        
        idx = idx[to_update]

        # Tag particles to update with the halo ID, mass and fractional radius
        part_halo_id[idx]       = halo_id[i]
        part_halo_mass[idx]     = halo_mass[i]
        part_halo_r_frac_2[idx] = r_frac_2[to_update]
        nr_assigned            += len(idx)

    nr_assigned_tot = comm.allreduce(nr_assigned)
    fraction_assigned = nr_assigned_tot / nr_particles_total
    message(f"Total particles assigned to halos = {nr_assigned_tot}")
    message(f"Fraction assigned = {fraction_assigned} (inc. duplicates due to halo overlap)")

    # Tidy up
    del halo_id
    del halo_pos
    del halo_radius
    del halo_mass
    del part_pos

    # Return r_frac=r/r200 for particles in halos and -1 for those not in halos
    in_halo = (part_halo_id >= 0)
    part_halo_r_frac = np.where(in_halo, np.sqrt(part_halo_r_frac_2), -1.0)
    del part_halo_r_frac_2

    # Replace any huge halo masses (i.e. particles not in any halo) with -1
    part_halo_mass[in_halo==False] = -1.0

    # Restore original particle ordering and return halo IDs etc
    message("Restoring particle order")
    order = psort.parallel_sort(part_index, return_index=True, comm=comm)
    del part_index
    psort.fetch_elements(part_halo_id, order, result=part_halo_id, comm=comm)
    psort.fetch_elements(part_halo_mass, order, result=part_halo_mass, comm=comm)
    psort.fetch_elements(part_halo_r_frac, order, result=part_halo_r_frac, comm=comm)

    return part_halo_id, part_halo_mass, part_halo_r_frac


def main(args):

    # Determine method to deal with overlapping halos
    overlap_method = overlap_methods[args.overlap_method]
    message(f"Halo overlap method: {args.overlap_method}")
    message(f"Halo radius definition: {args.soap_so_name}")

    # Read in position and radius for halos in the lightcone
    radius_name = f"{args.soap_so_name}/SORadius"
    mass_name = f"{args.soap_so_name}/TotalMass"
    halo_lightcone_data = read_lightcone_halo_positions_and_radii(args, radius_name, mass_name)

    # Locate the particle data
    type_z_range, all_particle_files = read_lightcone_index(args)

    # Generate filenames for the output:
    # These are the input filenames with the directory replaced with argument output_dir.
    output_filenames = []
    for input_filename in all_particle_files:
        dirname, filename = os.path.split(input_filename)
        output_filenames.append(os.path.join(args.output_dir, filename))

    # Open the input particle file set
    mf = phdf5.MultiFile(all_particle_files, comm=comm)

    # Loop over types to do
    create_files = True
    for ptype in type_z_range:
        
        message(f"Processing particle type {ptype}")

        # Read in positions of lightcone particles of this type
        message(f"Reading particles")
        part_pos = mf.read("Coordinates", group=ptype)

        # Record number of particles read from each file
        elements_per_file = mf.get_elements_per_file("Coordinates", group=ptype)

        # Rebalance particle load between MPI ranks
        nr_parts_per_rank_read = np.asarray(comm.allgather(part_pos.shape[0]), dtype=int)
        nr_parts_total = np.sum(nr_parts_per_rank_read)
        nr_parts_per_rank_balanced = np.zeros_like(nr_parts_per_rank_read)
        nr_av = (nr_parts_total // comm_size)
        nr_parts_per_rank_balanced[:] = nr_av
        nr_parts_per_rank_balanced[:nr_parts_total % comm_size] += 1
        assert np.sum(nr_parts_per_rank_balanced) == nr_parts_total
        part_pos = psort.repartition(part_pos, ndesired=nr_parts_per_rank_balanced, comm=comm)

        # Report load balancing
        max_nr_parts = comm.allreduce(part_pos.shape[0], op=MPI.MAX)
        min_nr_parts = comm.allreduce(part_pos.shape[0], op=MPI.MIN)
        message(f"No. of particles per rank min={min_nr_parts}, max={max_nr_parts}")

        # Assign group indexes to the particles
        message("Assigning group indexes")
        halo_id = halo_lightcone_data["IndexInHaloLightcone"]
        halo_pos = halo_lightcone_data["Pos_minpot"]
        halo_radius = halo_lightcone_data[radius_name]
        halo_mass = halo_lightcone_data[mass_name]
        part_halo_id, part_halo_mass, part_halo_r_frac = compute_particle_group_index(halo_id, halo_pos, halo_radius, halo_mass, part_pos, overlap_method)
        del part_pos
        del halo_id
        del halo_pos
        del halo_radius
        del halo_mass

        # Restore original partitioning of particles
        part_halo_id = psort.repartition(part_halo_id, ndesired=nr_parts_per_rank_read, comm=comm)
        part_halo_mass = psort.repartition(part_halo_mass, ndesired=nr_parts_per_rank_read, comm=comm)
        part_halo_r_frac = psort.repartition(part_halo_r_frac, ndesired=nr_parts_per_rank_read, comm=comm)

        # Write the output, appending to file if this is not the first particle type
        message(f"Writing output to {args.output_dir}")
        mode = "w" if create_files else "r+"
        datasets = {
            "IndexInHaloLightcone" : part_halo_id,
            "FractionalRadius" : part_halo_r_frac,
            "HaloMass" : part_halo_mass,
        }
        attributes = {
            "IndexInHaloLightcone" : halo_lightcone_data["ID"].attrs,
            "FractionalRadius" : halo_lightcone_data["ID"].attrs,
            "HaloMass" : halo_lightcone_data[mass_name].attrs,
        }
        mf.write(datasets, elements_per_file, output_filenames, mode,
                 group=ptype, attrs=attributes, gzip=6, shuffle=True)
                
        # Tidy up before reading next particle type
        del part_halo_id
        del part_halo_r_frac
        del part_halo_mass

        # Only need to create new output files for the first type
        create_files = False
        
    comm.barrier()

    # Discard reordered halo lightcone data
    del halo_lightcone_data
    del mf

    #
    # Now, for each particle in the lightcone we have the index in the halo
    # lightcone of the halo it belongs to.
    #
    # Next we re-read the halo lightcone to copy any extra halo properties we
    # want to store for each particle. This uses less memory than passing all
    # quantities through the calculation above.
    #
    message("Reading lightcone halo properties to copy to output particle files")
    halo_properties = (
        "Subhalo/ID",
        "Subhalo/SnapNum",
    )
    mf_in = phdf5.MultiFile(args.halo_lightcone_filenames, file_nr_attr=("Header", "NumberOfFiles"), comm=comm)
    halo_lightcone_data = mf_in.read(halo_properties, read_attributes=True)
    halo_lightcone_data["Subhalo/ID"] = halo_lightcone_data["Subhalo/ID"].astype(np.int64) # Avoid using unsigned int

    # Open the set of particle files to update
    mf_out = phdf5.MultiFile(output_filenames, comm=comm)

    # Loop over particle types to update
    for ptype in type_z_range:
   
        message(f"Reading halo index for particles of type {ptype}")
        halo_index = mf_out.read(f"{ptype}/IndexInHaloLightcone")
        elements_per_file = mf_out.get_elements_per_file(f"{ptype}/IndexInHaloLightcone")
        
        for prop_name in halo_properties:
            message(f"Pass through quantity {prop_name} for type {ptype}")
            in_halo = (halo_index >= 0)
            dtype = halo_lightcone_data[prop_name].dtype
            shape = (halo_index.shape[0],)+halo_lightcone_data[prop_name].shape[1:]
            prop_data = -np.ones(shape, dtype=dtype) # Set property=-1 if particle not in halo
            prop_data[in_halo,...] = psort.fetch_elements(halo_lightcone_data[prop_name], halo_index[in_halo], comm=comm)
            # Write the new dataset to the output files
            dataset_name = f"{ptype}/{prop_name.split('/')[-1]}"
            mf_out.write({dataset_name : prop_data}, elements_per_file, output_filenames, "r+",
                         attrs={dataset_name : halo_lightcone_data[prop_name].attrs},
                         gzip=6, shuffle=True)

    
if __name__ == "__main__":

    # Get command line arguments
    from virgo.mpi.util import MPIArgumentParser
    parser = MPIArgumentParser(description='Create lightcone halo catalogues.', comm=comm)
    parser.add_argument('lightcone_dir',  help='Directory with lightcone particle outputs')
    parser.add_argument('lightcone_base', help='Base name of the lightcone to use')
    parser.add_argument('halo_lightcone_filenames', help='Format string to generate halo lightcone filenames')
    parser.add_argument('soap_filenames', help='Format string to generate SOAP filenames')
    parser.add_argument('output_dir',     help='Where to write the output')
    parser.add_argument('--soap-so-name', type=str, default="SO/200_crit",
                        help='Name of SOAP group with the halo mass and radius, e.g. "SO/200_crit"')
    parser.add_argument('--overlap-method', type=str, default="fractional-radius", choices=list(overlap_methods),
                        help="How to assign particles which are in overlapping halos")
    args = parser.parse_args()

    message(f"Starting on {comm_size} MPI ranks")
    main(args)
    message("Done.")



#!/bin/env python
#
# This script uses the lightcone particle data to create a new HEALPix map
#

import healpy as hp
import numpy as np
import h5py

import lightcone_io.particle_reader as pr

# Specify one file from the spatially indexed lightcone particle data
input_filename = "/cosma8/data/dp004/jch/FLAMINGO/ScienceRuns/L1000N1800/DMO_FIDUCIAL/lightcones/lightcone0_particles/lightcone0_0000.0.hdf5"

# Where to write the new HEALPix map
output_filename = "/cosma8/data/dp004/jch/lightcone_map.hdf5"

# Open the lightcone
lightcone = pr.IndexedLightcone(input_filename)

# Create an empty HEALPix map
nside = 16384
npix = hp.pixelfunc.nside2npix(nside)
map_data = np.zeros(npix, dtype=float)

# Quantities to read in
property_names = ("Coordinates", "Masses", "ExpansionFactors")

# Redshift range to read in (we may get particles outside this range due to binning)
redshift_range = (0.0, 0.05)

# Part of the sky to do - set vector and radius to None for full sky.
# The pixel values outside this patch will be zero in the output map.
vector = (1.0, 0.0, 0.0)  # Vector pointing at a spot on the sky
radius = np.radians(10.0) # Angular radius around that spot

# Find out how many particles there will be
nr_particles_tot = lightcone["DM"].count_particles(redshift_range=redshift_range,
                                                   vector=vector, radius=radius,)
nr_particles_read = 0
nr_particles_added = 0
mass_added = 0.0

print("Total particles to read = %d" % nr_particles_tot)

# Loop over particles in the lightcone
for particle_chunk in lightcone["DM"].iterate_chunks(property_names=property_names,
                                                     redshift_range=redshift_range,
                                                     vector=vector, radius=radius,
                                                     max_particles=10*1024*1024):

    # Unpack position, mass and expansion factor arrays for this chunk
    pos = particle_chunk["Coordinates"]
    m   = particle_chunk["Masses"]
    z   = 1.0/particle_chunk["ExpansionFactors"]-1.0
    nr_particles_read += m.shape[0]

    # Filter out particles outside the required redshift range
    keep = (z >= redshift_range[0]) & (z < redshift_range[1])
    pos = pos[keep,:]
    m = m[keep]
    del z

    # Determine which pixels these particles belong to
    pos = pos.value # vec2pix can't handle unyt arrays, and we only need the direction from the vector anyway
    pixel = hp.pixelfunc.vec2pix(nside, pos[:,0], pos[:,1], pos[:,2])
    
    # Accumulate pixels to the map
    # This way is much slower because it updates all pixels in the map on each iteration
    #map_data += np.bincount(pixel, minlength=npix, weights=m)
    # Faster to use np.add.at to only update the affected pixels
    np.add.at(map_data, pixel, m.value)
    nr_particles_added += m.shape[0]
    mass_added += np.sum(m, dtype=float)

    # Report progress
    percent_done = (nr_particles_read/nr_particles_tot)*100.0
    print("Processed %d of %d particles (%.2f%%)" % (nr_particles_read, nr_particles_tot, percent_done))

print("Total number of particles added to map = %d" % nr_particles_added)
print("Total mass added to map = ", mass_added)
print("Sum over map pixels     = ", np.sum(map_data))

# Write out the result
with h5py.File(output_filename, "w") as outfile:
    outfile["MassMap"] = map_data
print("Wrote map to %s" % output_filename)



#!/bin/env python3

import sys

from mpi4py import MPI
comm = MPI.COMM_WORLD
comm_rank = comm.Get_rank()

from lightcone_io.particle_sorter import LightconeSorter


if __name__ == "__main__":

    from virgo.mpi.util import MPIArgumentParser
    parser = MPIArgumentParser(comm=comm, description="Reorder SWIFT particle lightcone outputs.")
    parser.add_argument("basedir", type=str, help="Lightcone output directory containing index files")
    parser.add_argument("basename", type=str, help="Name of the lightcone to process (e.g. lightcone0)")
    parser.add_argument("nr_redshift_bins", type=int, help="Number of redshift bins to use to bin particles")
    parser.add_argument("nside", type=int, help="HEALPpix map resolution to use to bin particles")
    parser.add_argument("outdir", type=str, help="Directory in which to write the output")
    parser.add_argument("--order", choices=["nest","ring"], default="nest", help="HEALPix pixel ordering scheme")
    parser.add_argument("--redshift-first", action="store_true", help="Sort by redshift then pixel (True) or pixel then redshift (False)")
    parser.add_argument("--lossy", action="store_true", help="Apply lossy compression to the output")
    parser.add_argument("--chunksize", type=int, default=0, help="HDF5 chunk size for output datasets")
    parser.add_argument("--types", type=str, default=None, help="Comma separated list of particle types to process (BH/DM/Gas/Stars/Neutrino)")
    args = parser.parse_args()

    # Seperate particle types list
    if args.types is not None:
        types = [t.strip() for t in args.types.split(",")]
    else:
        types = None
    
    if comm_rank == 0:
        if types is None:
            print("Processing all particle types")
        else:
            print("Types to process = ", types)
        print("Reading metadata")
    lightcone = LightconeSorter(args.basedir, args.basename, comm, types=types)

    if comm_rank == 0:
        print("Sorting lightcone")    
    lightcone.write_sorted_lightcone(args.outdir, args.nr_redshift_bins,
                                     args.nside, args.order, args.redshift_first,
                                     args.lossy, args.chunksize)
    comm.barrier()
    if comm_rank == 0:
        print("Done.")



#!/bin/env python

import collections.abc
from contextlib import nullcontext

import numpy as np
import re
import healpy as hp
import unyt

from .units import units_from_attributes
from .utils import LocalOrRemoteFile, SlicedDatasetReader


def merge_cells(cell_offset, cell_length):

    keep = cell_length > 0
    cell_length = cell_length[keep]
    cell_offset = cell_offset[keep]

    ncells = len(cell_length)
    j = 0
    for i in range(1,ncells):
        if (cell_offset[i] == cell_offset[j]+cell_length[j]):
            cell_length[j] += cell_length[i]
            cell_length[i] = 0
        else:
            j = i

    keep = cell_length > 0
    cell_length = cell_length[keep]
    cell_offset = cell_offset[keep]

    return cell_offset, cell_length


class IndexedLightconeParticleType(LocalOrRemoteFile):
    """
    Class to read a single particle type from a lightcone. This
    assumes that the lightcone particles have been sorted by redshift
    and healpix pixel index. We can then efficiently extract the
    particles in particular redshift ranges and regions of the sky.

    :param type_name: HDF5 group name for this particle type
    :type  type_name: str
    :param metadata: lightcone metadata read by :class:`ParticleLightcone`
    :type  metadata: dict
    :param units: system of units used in this output
    :type  units: dict
    :param filenames: names of the HDF5 files containing particle data
    :type  filenames: list of str
    :param extra_filenames: names of the HDF5 files containing additional data
    :type  extra_filenames: list of str or None
    :param remote_dir: remote directory containing the file, or None
    :type  remote_dir: hdfstream.RemoteDirectory, or None for local files
    """
    def __init__(self, type_name, metadata, index, units, filenames,
                 extra_filenames=None, remote_dir=None):
        self.set_directory(remote_dir)
        self.type_name = type_name
        self.metadata  = metadata
        self.index     = index
        self.units     = units
        self.filenames = filenames
        self.extra_filenames = extra_filenames
        self.comm      = None
        self.comm_rank = 0
        self.comm_size = 1

        # Find range of particles in each file
        first_particle_in_file = index["first_particle_in_file"]
        num_particles_in_file  = index["num_particles_in_file"]
        last_particle_in_file  = first_particle_in_file + num_particles_in_file - 1

        # Find first particle in each cell
        first_particle_in_cell = index["cell_offset"]

        # Find which cell the first and last particles in each file belong to
        self.index["first_cell_in_file"] = (
            np.searchsorted(first_particle_in_cell, first_particle_in_file, side="right") - 1)
        self.index["last_cell_in_file"] = (
            np.searchsorted(first_particle_in_cell, last_particle_in_file, side="right") - 1)

        # Find which quantities we have for this type
        properties = {}
        for filename in self.filenames:
            with self.open_file(filename) as infile:
                if type_name in infile:
                    for prop_name, dataset in infile[type_name].items():
                        if "a-scale exponent" in dataset.attrs:
                            shape = dataset.shape[1:]
                            dtype = dataset.dtype
                            if unyt is not None:
                                units = units_from_attributes(dataset)
                                properties[prop_name] = unyt.unyt_array(np.ndarray((0,)+shape, dtype=dataset.dtype), units)
                            else:
                                properties[prop_name] = np.ndarray((0,)+shape, dtype=dtype)
                            properties[prop_name].attrs = dict(dataset.attrs)
                    break
        self.properties = properties

        # Also check any extra files specified for additional quantities
        extra_properties = {}
        if extra_filenames is not None:
            assert len(filenames) == len(extra_filenames)
            for filename in self.extra_filenames:
                with self.open_file(filename) as infile:
                    if type_name in infile:
                        for prop_name, dataset in infile[type_name].items():
                            if "a-scale exponent" in dataset.attrs:
                                shape = dataset.shape[1:]
                                dtype = dataset.dtype
                                if unyt is not None:
                                    units = units_from_attributes(dataset)
                                    extra_properties[prop_name] = unyt.unyt_array(np.ndarray((0,)+shape, dtype=dataset.dtype), units)
                                else:
                                    extra_properties[prop_name] = np.ndarray((0,)+shape, dtype=dtype)
                                extra_properties[prop_name].attrs = dict(dataset.attrs)
                        break
        self.extra_properties = extra_properties

    def set_mpi_mode(self, comm):
        """
        Enable MPI mode for this output. In MPI mode each MPI rank reads a
        subset of the selected particles.

        :param comm: MPI communicator
        :type  comm: mpi4py.MPI.Comm
        """
        self.comm = comm
        self.comm_rank = comm.Get_rank()
        self.comm_size = comm.Get_size()

    def get_redshift_bins_in_range(self, redshift_min, redshift_max):
        """
        Return indexes of all redshift bins overlapping specified range

        :param redshift_min: minimum redshift
        :type  redshift_min: float
        :param redshift_max: maximum redshift
        :type  redshift_max: float

        :return: a numpy array of ints with the bin indexes
        :rtype: numpy.ndarray
        """
        redshift_bins = self.index["redshift_bins"]
        z_bin_min = redshift_bins[:-1]
        z_bin_max = redshift_bins[1:]
        read_bin = (z_bin_min <= redshift_max) & (z_bin_max >= redshift_min)
        z_bin_index = np.arange(len(z_bin_min), dtype=int)[read_bin]
        return z_bin_index

    def get_pixels_in_radius(self, vector, radius):
        """
        Return indexes of all healpix pixels within radius of vector

        :param vector: direction vector as an array of 3 floats
        :type  vector: numpy.ndarray
        :param radius: angular radius in radians
        :type  radius: float

        :return: a numpy array of ints with the pixel indexes
        :rtype: numpy.ndarray
        """
        nside = self.index["nside"]
        order = self.index["order"] if "order" in self.index else "ring"
        if isinstance(order, bytes):
            order = order.decode()
        if order == "nest":
            nest = True
        elif order == "ring":
            nest = False
        else:
            raise Exception("Invalid order parameter")
        pixels = hp.query_disc(nside, vector, radius, inclusive=True, nest=nest)
        pixels.sort()
        return pixels

    def get_cell_indexes_from_bins(self, redshift_bins, healpix_bins):
        """
        Given arrays of redshift and healpix bins to read, return
        indexes of all cells which must be read.

        :param redshift_bins: redshift bin indexes as array of ints
        :type  redshift_bins: numpy.ndarray
        :param healpix_bins: healpix pixel indexes as array of ints
        :type  healpix_bins: numpy.ndarray

        :return: a numpy array of ints with the cell indexes
        :rtype: numpy.ndarray
        """
        nr_pixels_to_read = len(healpix_bins)
        nr_redshifts_to_read = len(redshift_bins)
        nr_redshift_bins = len(self.index["redshift_bins"]) - 1
        nr_pixels = hp.pixelfunc.nside2npix(self.index["nside"])
        redshift_first = self.index["redshift_first"] if "redshift_first" in self.index else 0
        if redshift_first:
            redshift_bin_index = np.repeat(redshift_bins, nr_pixels_to_read)
            healpix_bin_index = np.tile(healpix_bins, nr_redshifts_to_read)
            cells_to_read = healpix_bin_index + (redshift_bin_index * nr_pixels)
        else:
            redshift_bin_index = np.tile(redshift_bins, nr_pixels_to_read)
            healpix_bin_index = np.repeat(healpix_bins, nr_redshifts_to_read)
            cells_to_read = redshift_bin_index + (healpix_bin_index * nr_redshift_bins)
        return cells_to_read

    def read_cells(self, property_names, cells_to_read):
        """
        Given an array of cells to read, read the specified properties
        from the lightcone particle files.

        :param property_names: list of properties to read
        :type  property_names: list of str
        :param cells_to_read: a numpy array of ints with the cell indexes
        :type  cells_to_read: numpy.ndarray

        :return: dict of ndarrays with the requested properties
        :rtype: dict
        """

        # Check property names all exist
        for name in property_names:
            if name not in self.properties and name not in self.extra_properties:
                raise KeyError(f"Can't find property: {name}")

        # Check if we need to read from the extra data files
        read_extra = False
        for name in property_names:
            if name not in self.properties and name in self.extra_properties:
                read_extra = True

        # Cell indexes should be in ascending order
        assert np.all(cells_to_read[1:] > cells_to_read[:-1])

        # Find number of particles to be read in
        nr_particles = np.sum(self.index["cell_length"][cells_to_read], dtype=np.int64)
        nr_cells_to_read = len(cells_to_read)

        # Dict to store the output arrays (don't know their data types yet)
        data = {name : None for name in property_names}
        offset = {}

        # Find range of particles in each file
        first_particle_in_file = self.index["first_particle_in_file"]
        num_particles_in_file  = self.index["num_particles_in_file"]

        # Find range of particles in each cell
        first_particle_in_cell = self.index["cell_offset"]
        num_particles_in_cell  = self.index["cell_length"]

        # Find range of cells in each file
        num_cells = self.index["num_cells"]
        first_cell_in_file = self.index["first_cell_in_file"]
        last_cell_in_file  = self.index["last_cell_in_file"]

        # Flag cells which we want to read
        read_cell = np.zeros(num_cells, dtype=bool)
        read_cell[cells_to_read] = True

        # Loop over files in the lightcone
        for current_file, filename in enumerate(self.filenames):

            # Locate file with extra datasets, if any
            if self.extra_filenames is not None:
                extra_filename = self.extra_filenames[current_file]
            else:
                extra_filename = None

            # Skip files with no particles
            if num_particles_in_file[current_file] == 0:
                continue

            # Find range of cells which have particles in this file
            first_cell = first_cell_in_file[current_file]
            last_cell  = last_cell_in_file[current_file]

            # Get start and end of each cell, relative to the start of this file
            cell_start = first_particle_in_cell[first_cell:last_cell+1] - first_particle_in_file[current_file]
            cell_stop  = cell_start + num_particles_in_cell[first_cell:last_cell+1]

            # Discard cells which will not be read
            keep = read_cell[first_cell:last_cell+1]
            cell_start = cell_start[keep]
            cell_stop  = cell_stop[keep]

            # Clip off any parts of cells which extend outside this file
            cell_start = np.clip(cell_start, a_min=0, a_max=None)
            cell_stop  = np.clip(cell_stop, a_min=None, a_max=num_particles_in_file[current_file])

            # Construct the sliced reader
            start = cell_start
            count = np.maximum(cell_stop-cell_start, 0)
            sliced_reader = SlicedDatasetReader(start, count)
            num_to_read = sliced_reader.count()

            # Skip files with no selected cells
            if sliced_reader.count() == 0:
                continue

            # Open the file(s)
            with self.open_file(filename) as infile, (self.open_file(extra_filename) if read_extra else nullcontext()) as extra_infile:

                # Loop over quantities to read
                for name in property_names:

                    # Find the dataset for this property
                    if name in self.properties:
                        dset = infile[self.type_name][name]
                    else:
                        dset = extra_infile[self.type_name][name]

                    # Create the output array, if we didn't already
                    if data[name] is None:
                        shape = list(dset.shape)
                        shape[0] = nr_particles
                        if unyt is not None:
                            units = units_from_attributes(dset)
                            data[name] = unyt.unyt_array(np.ndarray(shape, dtype=dset.dtype), units)
                        else:
                            data[name] = np.ndarray(shape, dtype=dset.dtype)
                        offset[name] = 0

                    # Read the cells
                    data[name][offset[name]:offset[name]+num_to_read,...] = sliced_reader.read(dset)
                    offset[name] += num_to_read

        # Check for the case where no particles were read
        if len(offset) == 0:
            # Nothing was read, so create empty arrays based on lightcone metadata
            for name in property_names:
                data[name] = np.empty_like(self.properties[name])
        else:
            # Read at least one cell, so check we read the expected number of particles
            for name in property_names:
                assert offset[name] == nr_particles

        return data

    def get_cell_indexes_from_vector_radius_redshift(self, vector, radius, redshift_range):
        """
        Given a line of sight vector, and angular radius and a redshift range,
        compute which cells we need to read.

        :param vector: direction vector as an array of 3 floats
        :type  vector: numpy.ndarray
        :param radius: angular radius in radians
        :type  radius: float
        :param redshift_range: redshift range to read
        :type  redshift_range: sequence of two floats [z_min, z_max]

        :return: a numpy array of ints with the cell indexes
        :rtype: numpy.ndarray
        """

        if redshift_range is not None:
            # Get redshift range to read
            redshift_min, redshift_max = [float(r) for r in redshift_range]
        else:
            # Read all redshifts
            redshift_min = self.index["redshift_bins"][0]
            redshift_max = self.index["redshift_bins"][-1]

        if vector is not None:
            # Select specified location on the sky
            if radius is None:
                raise ValueError("If specifying a radius must specify line of sight vector too")
            radius = float(radius)
            vector = np.asarray(vector, dtype=float)
        else:
            # Select full sky
            vector = np.asarray((1,0,0), dtype=float)
            radius = 2*np.pi

        # Find which redshift and healpix bins we should read in
        healpix_bins  = self.get_pixels_in_radius(vector, radius)
        redshift_bins = self.get_redshift_bins_in_range(redshift_min, redshift_max)
        cells_to_read = self.get_cell_indexes_from_bins(redshift_bins, healpix_bins)

        return cells_to_read

    def count_particles(self, vector=None, radius=None, redshift_range=None):
        """
        Return the number of particles in a region defined by a line of sight
        vector, an angular radius and a redshift range.

        :param vector: direction vector as an array of 3 floats
        :type  vector: numpy.ndarray
        :param radius: angular radius in radians
        :type  radius: float
        :param redshift_range: redshift range to read
        :type  redshift_range: sequence of two floats [z_min, z_max]

        :return: the number of particles selected
        :rtype: int
        """
        cells_to_read = self.get_cell_indexes_from_vector_radius_redshift(vector, radius, redshift_range)
        if self.comm is not None:
            cells_to_read = self._split_cells_by_mpi_rank(cells_to_read)
        cell_size = self.index["cell_length"][cells_to_read]
        return np.sum(cell_size)

    def iterate_chunks(self, property_names, vector=None, radius=None, redshift_range=None,
                       max_particles=1048576):
        """
        Iterate over chunks of particles in a region defined by a line of sight
        vector, an angular radius and a redshift range. This can be used to
        avoid reading all of the particles into memory at once. Yields dicts
        of arrays where the keys are the requested particle property names and
        the values are the particle property values.

        :param vector: direction vector as an array of 3 floats
        :type  vector: numpy.ndarray
        :param radius: angular radius in radians
        :type  radius: float
        :param redshift_range: redshift range to read
        :type  redshift_range: sequence of two floats [z_min, z_max]
        :param max_particles: maximum number of particles to return in one chunk
        :type  max_particles: int

        :yield: dict of ndarray with the particle properties for each chunk
        :rtype: dict
        """
        cells_to_read = self.get_cell_indexes_from_vector_radius_redshift(vector, radius, redshift_range)
        if self.comm is not None:
            cells_to_read = self._split_cells_by_mpi_rank(cells_to_read)
        nr_cells = len(cells_to_read)

        # Find the sizes of the selected cells
        cell_size = self.index["cell_length"][cells_to_read]

        # Loop over and yield chunks of cells containing up to max_particles
        i1 = 0
        nr_yielded = 0
        while i1 < nr_cells:

            # Will always read at least one cell
            i2 = i1 + 1
            nr_to_read = cell_size[i1]

            # Determine how many more cells to read
            while i2 < nr_cells and nr_to_read+cell_size[i2] <= max_particles:
                nr_to_read += cell_size[i2]
                i2 += 1

            # Read and return the cell(s), if they contain any particles
            if nr_to_read > 0:
                yield self.read_cells(property_names, cells_to_read[i1:i2])
                nr_yielded += 1

            # Advance to the next set of cells
            i1 = i2

        # If we didn't read anything, return a set of empty arrays
        if nr_yielded == 0:
            yield self.read_cells(property_names, cells_to_read[1:0])

    def _split_cells_by_mpi_rank(self, cells_to_read):

        # Find number of cells to read on each MPI rank
        nr_per_rank = np.zeros(self.comm_size, dtype=int)
        nr_per_rank[...] = len(cells_to_read) // self.comm_size
        nr_per_rank[:len(cells_to_read) % self.comm_size] += 1
        assert sum(nr_per_rank) == len(cells_to_read)

        # Find offset to first cell on each rank
        offset_on_rank = np.cumsum(nr_per_rank) - nr_per_rank

        # Return range to read on this rank
        i1 = offset_on_rank[self.comm_rank]
        i2 = i1 + nr_per_rank[self.comm_rank]
        return cells_to_read[i1:i2]

    def read(self, property_names, vector=None, radius=None, redshift_range=None):
        """
        Read all particles in a region defined by a line of sight vector,
        an angular radius and a redshift range. Returns a dict of arrays where
        the keys are the requested particle property names and the values are
        the particle property values. Always returns complete cells, so it may
        return particles outside the requested region.

        :param property_names: list of properties to read
        :type  property_names: list of str
        :param vector: direction vector as an array of 3 floats
        :type  vector: numpy.ndarray
        :param radius: angular radius in radians
        :type  radius: float
        :param redshift_range: redshift range to read
        :type  redshift_range: sequence of two floats [z_min, z_max]

        :return: dict of ndarray with the particle properties
        :rtype: dict
        """
        # Read data for the selected cells
        cells_to_read = self.get_cell_indexes_from_vector_radius_redshift(vector, radius, redshift_range)
        if self.comm is not None:
            cells_to_read = self._split_cells_by_mpi_rank(cells_to_read)
        data = self.read_cells(property_names, cells_to_read)
        return data

    def read_exact(self, property_names, vector=None, radius=None, redshift_range=None):
        """
        Read all particles in a region defined by a line of sight vector,
        an angular radius and a redshift range. Returns a dict of arrays where
        the keys are the requested particle property names and the values are
        the particle property values. This version filters out any particles
        outside the requested region.

        :param property_names: list of properties to read
        :type  property_names: list of str
        :param vector: direction vector as an array of 3 floats
        :type  vector: numpy.ndarray
        :param radius: angular radius in radians
        :type  radius: float
        :param redshift_range: redshift range to read
        :type  redshift_range: sequence of two floats [z_min, z_max]

        :return: dict of ndarray with the particle properties
        :rtype: dict
        """
        if vector is not None:
            vector = np.asarray(vector, dtype=float)

        # Filtering on angle is only implemented for radii < 90 degrees
        if radius is not None and radius > 0.5*np.pi:
            raise Exception("Angular radius must be <= pi/2 radians")

        # Make a dict to store the output.
        data = {name : [] for name in property_names}

        # Ensure we're reading the expansion factors if we need them
        keep_expansion = "ExpansionFactors" in property_names
        if redshift_range is not None and "ExpansionFactors" not in property_names:
            property_names = list(property_names) + ["ExpansionFactors",]

        # Ensure we're reading the coordinates if we need them
        keep_coordinates = "Coordinates" in property_names
        if radius is not None and "Coordinates" not in property_names:
            property_names = list(property_names) + ["Coordinates",]

        # Then we'll read the data in chunks and filter one chunk at a time
        max_particles = 10*1024*1024 # Use large-ish reads for efficiency
        for chunk in self.iterate_chunks(property_names, vector, radius, redshift_range, max_particles):

            # Find number of particles in this chunk
            nr_part = chunk[property_names[0]].shape[0]

            # Filter particles on redshift
            if redshift_range is not None:
                a_min = 1.0/(1.0+redshift_range[1])
                a_max = 1.0/(1.0+redshift_range[0])
                keep = (chunk["ExpansionFactors"]>=a_min) & (chunk["ExpansionFactors"]<=a_max)
            else:
                keep = np.ones(nr_part, dtype=bool)

            # Filter particles on radius
            if radius is not None:
                # Get normalized view direction vector
                norm_vector = np.asarray(vector, dtype=float)
                norm_vector = norm_vector / np.sqrt(np.sum(norm_vector**2))
                # Get 3D distance squared for each particle
                pos = chunk["Coordinates"]
                r2 = np.sum(pos**2, axis=1)
                # Find angle between view vector and each particle
                cos2_theta = (np.sum(vector[None,:]*pos, axis=1)**2)/r2
                keep = keep & (cos2_theta >= np.cos(radius)**2)
                del pos
                del r2
                del cos2_theta

            # Discard expansion factors if we were not asked to read them
            if keep_expansion == False and "ExpansionFactors" in chunk:
                del chunk["ExpansionFactors"]

            # Discard coordinates if we were not asked to read them
            if keep_coordinates == False and "Coordinates" in chunk:
                del chunk["Coordinates"]

            # Store all remaining properties
            for name in data:
                data[name].append(chunk[name][keep,...])
            del keep

        # Merge chunks
        for name in data:
            data[name] = np.concatenate(data[name])

        return data


class ParticleLightcone(collections.abc.Mapping, LocalOrRemoteFile):
    """
    Class used to read particle lightcones. This is a dict-like container for
    the :class:`IndexedLightconeParticleType` instances which represent the
    different particle types in this lightcone (gas, DM, stars etc).

    Particle lightcones are opened by specifying the name of any one of the
    particle data files.

    If an MPI communicator is specified, then reads are parallelized over MPI
    ranks in the communicator.

    :param filename: name of one of the lightcone particle files
    :type  filename: str
    :param comm: MPI communicator
    :type  comm: mpi4py.MPI.Comm, or None
    :param extra_filename: name of a file with extra particle properties
    :type  extra_filename: str, or None
    :param remote_dir: remote directory containing the file, or None
    :type  remote_dir: hdfstream.RemoteDirectory, or None for local files
    """
    def __init__(self, fname, comm=None, extra_filename=None, remote_dir=None):
        self.set_directory(remote_dir)
        if comm is not None:
            comm_rank = comm.Get_rank()
            comm_size = comm.Get_size()
        else:
            comm_rank = 0
            comm_size = 1

        # In MPI mode rank 0 reads the metadata and then we broadcast it
        if comm_rank == 0:

            # Store base name so we can find the other files
            m = re.match(r"(.*)\.[0-9]+\.hdf5", fname)
            if m is not None:
                basename = m.group(1)
            else:
                raise IOError("Unable to extract base name from filename: %s" % fname)

            # Repeat for extra data files, if any
            if extra_filename is not None:
                m = re.match(r"(.*)\.[0-9]+\.hdf5", extra_filename)
                if m is not None:
                    extra_basename = m.group(1)
                else:
                    raise IOError("Unable to extract base name from filename: %s" % fname)

            with self.open_file(fname) as infile:

                # Check that this file has indexing info: this class cannot be used to read the
                # un-sorted lightcones output directly by Swift.
                if "Cells" not in infile:
                    raise RuntimeError("This does not appear to be an indexed particle lightcone output")

                # Read metadata from the specified file
                metadata = {}
                for name in infile["Lightcone"].attrs:
                    metadata[name] = infile["Lightcone"].attrs[name]

                # Read units information
                units = {}
                units["Current"]     = infile["Units"].attrs["Unit current in cgs (U_I)"][0]
                units["Length"]      = infile["Units"].attrs["Unit length in cgs (U_L)"][0]
                units["Mass"]        = infile["Units"].attrs["Unit mass in cgs (U_M)"][0]
                units["Temperature"] = infile["Units"].attrs["Unit temperature in cgs (U_T)"][0]
                units["Time"]        = infile["Units"].attrs["Unit time in cgs (U_t)"][0]

                # Find names of all of the lightcone files
                filenames = []
                for i in range(int(metadata["nr_mpi_ranks"][0])):
                    filenames.append("%s.%d.hdf5" % (basename, i))

                # Find names of extra files
                if extra_filename is not None:
                    extra_filenames = []
                    for i in range(int(metadata["nr_mpi_ranks"][0])):
                        extra_filenames.append("%s.%d.hdf5" % (extra_basename, i))
                else:
                    extra_filenames = None

                # Determine particle types present
                self.particle_types = {}
                for type_name in infile["Cells"]:
                    index = {}
                    for name in infile["Cells"][type_name]:
                        index[name] = infile["Cells"][type_name][name][()]
                    self.particle_types[type_name] = IndexedLightconeParticleType(type_name, metadata, index, units, filenames,
                                                                                  extra_filenames, remote_dir)
        else:
            self.particle_types = None

        if comm is not None:
            self.particle_types = comm.bcast(self.particle_types)
            for pt in self.particle_types.values():
                pt.set_mpi_mode(comm)

    def __getitem__(self, key):
        return self.particle_types[key]

    def __iter__(self):
        for name in self.particle_types:
            yield name

    def __len__(self):
        return len(self.particle_types)

# Alias for backwards compatibility
IndexedLightcone = ParticleLightcone



#!/bin/env python

import os
import sys
import numpy as np
import h5py
import argparse
from mpi4py import MPI
import unyt
import healpy as hp

import virgo.util.match as match
import virgo.mpi.parallel_hdf5 as phdf5
import virgo.mpi.parallel_sort as psort
from virgo.mpi.util import MPIArgumentParser
import virgo.formats.swift

import lightcone_io.particle_reader as pr
import lightcone_io.halo_catalogue as hc
import lightcone_io.choose_bh_tracer as ct

comm = MPI.COMM_WORLD
comm_rank = comm.Get_rank()
comm_size = comm.Get_size()

# Special most bound black hole ID for halos with no black holes.
# Must not appear as a particle ID.
NULL_BH_ID = np.iinfo(np.int64).max

# Particle type names used in the lightcone
part_type_names = [
    "Gas",
    "DM",
    None,
    None,
    "Stars",
    "BH",
    "Neutrino",
    ]

# HDF5 chunk size for the output
chunk_size = 1024*1024

def redshift_ranges(snap_z, first_snap, last_snap):
    """
    Given a dict of snapshot redshifts, assign a redshift range to each
    snapshot which extends half way to adjacent snasphots.
    """
    z1 = {}
    z2 = {}
    for snap_nr in range(first_snap, last_snap+1):
        if snap_nr == first_snap:
            z2[snap_nr] = snap_z[first_snap]
        else:
            z2[snap_nr] = 0.5*(snap_z[snap_nr-1]+snap_z[snap_nr])
        if snap_nr == last_snap:
            z1[snap_nr] = snap_z[last_snap]
        else:
            z1[snap_nr] = 0.5*(snap_z[snap_nr]+snap_z[snap_nr+1])
    return z1, z2


def attributes_from_units(units):
    """
    Given a unyt.Unit object, generate SWIFT dataset attributes

    units: the Unit object

    Returns a dict with the attributes
    """
    attrs = {}

    # Get CGS conversion factor. Note that this is the conversion to physical units,
    # because unyt multiplies out the dimensionless a factor.
    cgs_factor, offset = units.get_conversion_factor(units.get_cgs_equivalent())

    # Get a exponent
    a_unit = unyt.Unit("a", registry=units.registry)
    a_exponent = units.expr.as_powers_dict()[a_unit.expr]
    a_val = a_unit.base_value

    # Get h exponent
    h_unit = unyt.Unit("h", registry=units.registry)
    h_exponent = units.expr.as_powers_dict()[h_unit.expr]
    h_val = h_unit.base_value

    # Find the power associated with each dimension
    powers = units.get_mks_equivalent().dimensions.as_powers_dict()

    # Set the attributes
    attrs["Conversion factor to CGS (not including cosmological corrections)"] = [
        float(cgs_factor / (a_val ** a_exponent) / (h_val ** h_exponent))
    ]
    attrs["Conversion factor to physical CGS (including cosmological corrections)"] = [
        float(cgs_factor)
    ]
    attrs["U_I exponent"] = [float(powers[unyt.dimensions.current_mks])]
    attrs["U_L exponent"] = [float(powers[unyt.dimensions.length])]
    attrs["U_M exponent"] = [float(powers[unyt.dimensions.mass])]
    attrs["U_T exponent"] = [float(powers[unyt.dimensions.temperature])]
    attrs["U_t exponent"] = [float(powers[unyt.dimensions.time])]
    attrs["a-scale exponent"] = [float(a_exponent)]
    attrs["h-scale exponent"] = [float(h_exponent)]

    return attrs


def message(m):
    if comm_rank == 0:
        print(m)


def drop_a_from_comoving_length(arr):
    """
    Convert an array of lengths to comoving snapshot length units and drop
    the a factor from the units without changing the value.

    Coordinates in the lightcone are comoving but the expansion factor varies
    so we can't include it explicitly in the units as we might in snapshots.
    Here we convert to units a*snap_length and reinterpret the numbers as
    just snap_length.
    """
    reg = arr.units.registry
    return arr.to("a*snap_length").value * unyt.Unit("snap_length", registry=reg)


def match_black_holes(args):

    message(f"Starting on {comm_size} MPI ranks")

    # Determine quantities to read from the halo catalogue
    to_read = ["InputHalos/HaloCentre", "InputHalos/HaloCatalogueIndex"]
    if args.pass_through is not None:
        for prop_name in args.pass_through.split(","):
            to_read.append(prop_name)

    # Open the lightcone particle output in MPI mode
    filename = f"{args.lightcone_dir}/{args.lightcone_base}_particles/{args.lightcone_base}_0000.0.hdf5"
    lightcone = pr.IndexedLightcone(filename, comm=comm)

    if comm_rank == 0:
        # Get simulation box size and unit system from a snapshot file.
        # HBT output specifies how to convert its chosen units to Mpc and Msolar so
        # we need the snapshot unit system to interpret the catalogues using SWIFT's
        # physical constants for consistency.
        filename = args.snapshot_format.format(snap_nr=args.last_sim_snap, file_nr=0)
        with h5py.File(filename, "r") as infile:
            boxsize_no_units = infile["Header"].attrs["BoxSize"][0]
            swift_unit_registry = virgo.formats.swift.soap_unit_registry_from_snapshot(infile)
        # Read all snapshot redshifts
        redshifts = {}
        for snap_nr in range(args.first_sim_snap, args.last_sim_snap+1):
            filename = args.snapshot_format.format(snap_nr=snap_nr, file_nr=0)
            with h5py.File(filename, "r") as infile:
                redshifts[snap_nr] = float(infile["Header"].attrs["Redshift"][0])
    else:
        boxsize_no_units = None
        swift_unit_registry = None
    boxsize_no_units, swift_unit_registry = comm.bcast((boxsize_no_units, swift_unit_registry))

    # Assign units to the boxsize. Box size is comoving but we deliberately
    # omit the a factor here.
    boxsize = boxsize_no_units * unyt.Unit("snap_length", registry=swift_unit_registry)

    # Open the halo catalogue
    if args.halo_type == "SOAP":
        halo_cat = hc.SOAPCatalogue(args.halo_format, args.first_sim_snap, args.last_sim_snap)
    elif args.halo_type == "HBTplus":
        halo_cat = hc.HBTplusCatalogue(args.halo_format, args.snapshot_format, args.first_sim_snap, args.last_sim_snap)
    else:
        raise ValueError("Unrecognized value for --halo-type option")

    # Get the redshift range associated with each snapshot
    all_z1, all_z2 = redshift_ranges(halo_cat.redshift, args.first_sim_snap, args.last_sim_snap)

    # Loop over snapshots
    halos_so_far = 0
    membership_cache = {}
    for snap_nr in range(args.end_snap, args.start_snap-1, -1):

        # Read halos at this snapshot
        halo_data = halo_cat.read(snap_nr, to_read)

        # Count halos
        nr_halos_in_slice = len(halo_data["InputHalos/HaloCatalogueIndex"])
        nr_halos_in_slice_all = comm.allreduce(nr_halos_in_slice)

        # Choose the tracer BH particle to use for each object.
        # Returns ID and position of the selected BH particle.
        part_type = f"PartType{args.part_type}"
        message(f"  Choosing tracer particles for snapshot {snap_nr} using {part_type}")
        tracer_id, tracer_pos = ct.choose_bh_tracer(halo_data["InputHalos/HaloCatalogueIndex"],
                                                    snap_nr, args.last_sim_snap, args.snapshot_format,
                                                    args.membership_format, membership_cache,
                                                    part_type, NULL_BH_ID)
        tracer_pos = drop_a_from_comoving_length(tracer_pos)

        # Each snapshot populates a redshift range which reaches half way to adjacent snapshots
        # (range is truncated for the first and last snapshots)
        z_snap = halo_cat.redshift[snap_nr]
        z1 = all_z1[snap_nr]
        z2 = all_z2[snap_nr]
        message(f"  Using {nr_halos_in_slice_all} halos at z={z_snap:.3f} to populate range z={z1:.3f} to z={z2:.3f}")

        # Read in the lightcone BH particle positions and IDs in this redshift range
        lightcone_props = ("Coordinates", "ParticleIDs", "ExpansionFactors")
        part_type_name = part_type_names[args.part_type]
        particle_data = lightcone[part_type_name].read_exact(lightcone_props, redshift_range=(z1,z2))
        if np.any(particle_data["ParticleIDs"] == NULL_BH_ID):
            raise RuntimeError("Found a BH particle with ID=NULL_BH_ID!")
        nr_parts = len(particle_data["ParticleIDs"])
        nr_parts_all = comm.allreduce(nr_parts)
        message(f"  Read in {nr_parts_all} lightcone BH particles for this redshift range")

        # Try to match BH particles to the halo most bound BH IDs.
        # There may be multiple particles matching each halo due to the periodicity of the box.
        # Since halos with no black hole have tracer_id=NULL_BH_ID and this value never appears
        # in the particle data, every match will become a halo in the output catalogue.
        halo_index = psort.parallel_match(particle_data["ParticleIDs"], tracer_id, comm=comm)
        matched = halo_index>=0
        nr_matched = np.sum(matched)
        nr_matched_all = comm.allreduce(nr_matched)
        halos_so_far += nr_matched
        halo_index = halo_index[matched]
        if nr_parts_all > 0:
            pc_matched = 100.0*(nr_matched_all/nr_parts_all)
        else:
            pc_matched = 100.0
        message(f"  Matched {nr_matched_all} BH particles in this slice ({pc_matched:.2f}%)")

        # Create the output halo catalogue for this redshift slice:
        # For each matched BH particle in the lightcone, we fetch the properties of the halo
        # it was matched with.
        halo_slice = {}
        for name in sorted(halo_data):
            halo_slice[name] = psort.fetch_elements(halo_data[name], halo_index, comm=comm)
        message(f"  Found halo properties for this slice")

        # Compute the position of each halo in the output:
        #
        # We know:
        #  - the position in the lightcone of the most bound black hole particle
        #  - the position in the snapshot of the most bound black hole particle
        #  - the position in the snapshot of the halo's potential minimum
        #
        # We want to compute the position in the lightcone of the potential minimum.
        #
        registry = tracer_pos.units.registry
        length_unit = unyt.Unit("snap_length", registry=registry)

        # Position of the matched BH particle in the lightcone particle output.
        # Has one entry for each BH in the lightcone which matched a halo.
        bh_pos_in_lightcone = particle_data["Coordinates"][matched,...].to(length_unit)

        # Position of the selected tracer BH, taken from the snapshot.
        bh_pos_in_snapshot = psort.fetch_elements(tracer_pos, halo_index, comm=comm).to(length_unit)

        # Position of the matched halo from the halo finder:
        # Note that the units include an a factor, which we need to remove
        halo_pos_in_snapshot = drop_a_from_comoving_length(halo_slice["InputHalos/HaloCentre"]).to(length_unit)

        # Vector from the tracer BH to the potential minimum - may need box wrapping
        bh_to_minpot_vector = halo_pos_in_snapshot - bh_pos_in_snapshot
        bh_to_minpot_vector = ((bh_to_minpot_vector+0.5*boxsize) % boxsize) - 0.5*boxsize

        # Compute halo potential minimum position in lightcone
        halo_pos_in_lightcone = bh_pos_in_lightcone + bh_to_minpot_vector

        # Report largest offset between BH and potential minimum
        max_offset = ct.distributed_amax(bh_to_minpot_vector.flatten(), comm)
        if comm_rank == 0 and max_offset is not None:
            message(f"  Maximum minpot/bh offset = {max_offset} (SWIFT length units, comoving)")

        # Add the position and redshift in the lightcone to the output catalogue
        halo_slice["Lightcone/HaloCentre"] = halo_pos_in_lightcone
        #halo_slice["Lightcone/TracerPosition"] = bh_pos_in_lightcone
        halo_slice["Lightcone/Redshift"] = unyt.unyt_array(1.0/particle_data["ExpansionFactors"][matched]-1.0,
                                                           units="dimensionless", registry=registry)
        # Add snapshot number to the output catalogue
        snap_nr_arr = np.ones(len(halo_slice["Lightcone/Redshift"]), dtype=int)*snap_nr
        halo_slice["Lightcone/SnapshotNumber"] = unyt.unyt_array(snap_nr_arr, dtype=int, units="dimensionless", registry=registry)

        message(f"  Computed potential minimum position in lightcone")

        # Now we need to sort all of the halos by their healpix pixel index
        vectors = halo_pos_in_lightcone.ndview # healpy can't handle unyt arrays
        pixel_index = hp.pixelfunc.vec2pix(args.nside, vectors[:,0], vectors[:,1],
                                           vectors[:,2], nest=(args.order=="nest"))
        del vectors
        message(f"  Computed pixel index for each halo")

        order = psort.parallel_sort(pixel_index, return_index=True, comm=comm)
        message(f"  Computed sorting order for the halos")

        # Count how many halos there are in each pixel
        npix = hp.pixelfunc.nside2npix(args.nside)
        halos_per_pixel = psort.parallel_bincount(pixel_index, minlength=npix, comm=comm)
        del pixel_index
        message(f"  Computed number of halos per pixel")

        # Reorder the halo properties
        for name in sorted(halo_slice):
            halo_slice[name] = psort.fetch_elements(halo_slice[name], order, comm=comm)
            message(f"      Re-ordered halo property: {name}")

        # Write out the halo catalogue for this snapshot
        output_filename = f"{args.output_dir}/lightcone_halos_{snap_nr:04d}.hdf5"
        outfile = h5py.File(output_filename, "w", driver="mpio", comm=comm, libver="v108")
        for name in sorted(halo_slice):
            # Ensure the group exists
            outfile.require_group(os.path.dirname(name))
            # Write the data
            dset = phdf5.collective_write(outfile, name, halo_slice[name], gzip=6, chunk=chunk_size, comm=comm)
            # Write units
            attrs = attributes_from_units(halo_slice[name].units)
            for attr_name, attr_val in attrs.items():
                dset.attrs[attr_name] = attr_val
            # Write description
            dset.attrs["Description"] = halo_cat.description[name]

        # Write the indexing information to the file
        index_group = outfile.require_group("Index")
        index_group.attrs["nside"] = args.nside
        index_group.attrs["order"] = args.order
        phdf5.collective_write(index_group, "NumHalosPerPixel", halos_per_pixel, gzip=6, chunk=chunk_size, comm=comm)

        # Also write the offset to the first halo in each pixel. Note that halos_per_pixel
        # is distributed over all MPI ranks.
        total_halos_this_rank = np.sum(halos_per_pixel, dtype=np.int64)
        total_halos_prev_ranks = comm.scan(total_halos_this_rank) - total_halos_this_rank
        first_halo_in_pixel = np.cumsum(halos_per_pixel, dtype=np.int64) - halos_per_pixel + total_halos_prev_ranks
        phdf5.collective_write(index_group, "FirstHaloInPixel", first_halo_in_pixel, gzip=6, chunk=chunk_size, comm=comm)

        # Write out the range of redshifts associated with each snapshot
        snap_index = np.arange(args.first_sim_snap, args.last_sim_snap+1)
        z_min = np.asarray([all_z1[sn] for sn in snap_index], dtype=float)
        z_max = np.asarray([all_z2[sn] for sn in snap_index], dtype=float)
        z_snap = np.asarray([halo_cat.redshift[sn] for sn in snap_index], dtype=float)
        snap_group = outfile.create_group("Snapshots")
        snap_group.attrs["snapshot_numbers"] = snap_index
        snap_group.attrs["snapshot_redshifts"] = z_snap
        snap_group.attrs["minimum_redshifts"] = z_min
        snap_group.attrs["maximum_redshifts"] = z_max
        snap_group.attrs["first_snapshot_number"] = args.first_sim_snap
        snap_group.attrs["last_snapshot_number"] = args.last_sim_snap
        snap_group.attrs["this_snapshot_number"] = snap_nr

        # Write a list of property names to the file
        outfile["Lightcone"].attrs["property_names"] = list(halo_slice.keys())

        # Correct a-exponent of the lightcone positions (they're comoving)
        outfile["Lightcone/HaloCentre"].attrs["a-scale exponent"] = (1.0,)
        outfile.close()
        message(f"  Wrote file: {output_filename}")

        # Tidy up particle arrays before we read the next slice
        del particle_data

    comm.barrier()
    message("All redshift ranges done.")


def run():

    parser = MPIArgumentParser(comm, description='Create lightcone halo catalogues.')
    parser.add_argument('halo_format', help='Format string for halo catalogue filenames (using {snap_nr}, {file_nr})')
    parser.add_argument('first_sim_snap', type=int, help='Index of the first snapshot of the simulation')
    parser.add_argument('last_sim_snap', type=int, help='Index of the last snapshot of the simulation')
    parser.add_argument('start_snap', type=int, help='Index of the first snapshot to process')
    parser.add_argument('end_snap', type=int, help='Index of the last snapshot to process')
    parser.add_argument('lightcone_dir',  help='Directory with lightcone particle outputs')
    parser.add_argument('lightcone_base', help='Base name of the lightcone to use')
    parser.add_argument('snapshot_format',  help='Format string for snapshot filenames (e.g. "snap_{snap_nr:04d}.{file_nr}.hdf5")')
    parser.add_argument('membership_format', help='Format string for group membership filenames')
    parser.add_argument('output_dir',     help='Where to write the output')
    parser.add_argument('--pass-through', default=None, help='Comma separated list of SOAP properties to pass through')
    parser.add_argument('--halo-type', choices=("SOAP","HBTplus"), default="SOAP", help='Input halo catalogue type')
    parser.add_argument('--part-type', type=int, default=5, help='Particle type to use for placing lightcone halos')
    parser.add_argument("--nside", type=int, default=16, help="HEALPpix map resolution to use to bin halos in the output")
    parser.add_argument("--order", choices=["nest","ring"], default="nest", help="HEALPix pixel ordering scheme")
    args = parser.parse_args()
    match_black_holes(args)


if __name__ == "__main__":
    run()



#!/bin/env python

import subprocess
import sys
import os
import shutil
import numpy as np
import h5py
import yaml
from mpi4py import MPI
from mpi4py.futures import MPICommExecutor

comm = MPI.COMM_WORLD
comm_rank = comm.Get_rank()
comm_size = comm.Get_size()


def correct_shell_file(indir, basename, outdir, shell_nr, z_central, neutrino_density):

    # Generate shell input and output file names
    infile = ("%s/%s_shells/shell_%d/%s.shell_%d.0.hdf5" %
              (indir, basename, shell_nr, basename, shell_nr))
    outfile = ("%s/%s_shells/shell_%d/%s.shell_%d.0.hdf5" %
               (outdir, basename, shell_nr, basename, shell_nr))

    # Create the output directory
    try:
        os.makedirs(os.path.dirname(outfile))
    except OSError:
        pass

    # Check that the input file exists
    if not(os.path.exists(infile)):
        raise RuntimeError(f"Input file {infile} does not exist!")

    # Copy the file
    #shutil.copyfile(infile, outfile)
    #subprocess.run(["cp", infile, outfile], check=True)

    # Move the file
    subprocess.run(["mv", infile, outfile], check=True)

    #
    # Now need to apply corrections
    #
    f = h5py.File(outfile, "r+")

    # DopplerB needs to be multiplied by (1+z)
    if "DopplerB" in f:
        dset = f["DopplerB"]
        if "Correction factor" not in dset.attrs:
            fac = (1.0+z_central[shell_nr])
            dset.attrs["Correction applied"] = "Multiplied by 1+(z_min+z_max)/2 to compensate for bug in original output"
            dset.attrs["Correction factor"] = fac
            dset.attrs["Central redshift assumed for correction"] = z_central[shell_nr]
            nchunk=100*1024*1024
            ntot = dset.shape[0]
            for offset in range(0, ntot, nchunk):
                n = min(nchunk, ntot-offset)
                data = dset[offset:offset+n,...]
                data *= fac
                dset[offset:offset+n,...] = data
        else:
            print(f"  Skipping DopplerB correction for shell {shell_nr} of {basename}")

    # DM needs to be multiplied by 1/(1+z)
    if "DM" in f:
        dset = f["DM"]
        if "Correction factor" not in dset.attrs:
            fac = 1.0/(1.0+z_central[shell_nr])
            dset.attrs["Correction applied"] = "Multiplied by 1/(1+(z_min+z_max)/2) to compensate for bug in original output"
            dset.attrs["Correction factor"] = fac
            dset.attrs["Central redshift assumed for correction"] = z_central[shell_nr]
            nchunk=100*1024*1024
            ntot = dset.shape[0]
            for offset in range(0, ntot, nchunk):
                n = min(nchunk, ntot-offset)
                data = dset[offset:offset+n,...]
                data *= fac
                dset[offset:offset+n,...] = data
        else:
            print(f"  Skipping DM correction for shell {shell_nr} of {basename}")


    # Neutrino mass maps:
    # Need to subtract 3*(mean neutrino density) to get actual neutrino density.
    # Will need to correct the neutrino mass and total mass maps. The mean
    # neutrino density can be extracted from the snapshot file.
    for name in ("NeutrinoMass", "TotalMass"):
        if name in f:
            dset = f[name]
            if "Correction offset" not in dset.attrs:
                # Find the volume of a pixel
                r1 = dset.attrs["comoving_inner_radius"][0]
                r2 = dset.attrs["comoving_outer_radius"][0]
                npix = dset.attrs["number_of_pixels"]
                pixel_volume = (4./3.)*np.pi*(r2**3-r1**3)/npix
                # Compute offset to apply to the map
                mass_per_pixel = -3.0*(neutrino_density*pixel_volume)
                # Update attributes with correction to apply
                dset.attrs["Correction applied"] = ["Subtracted 3 times the mean neutrino mass per pixel to fix bug in original output"]
                dset.attrs["Correction offset"] = mass_per_pixel
                # Apply  the correction to the neutrino mass
                nchunk=100*1024*1024
                ntot = dset.shape[0]
                for offset in range(0, ntot, nchunk):
                    n = min(nchunk, ntot-offset)
                    data = dset[offset:offset+n,...]
                    data += mass_per_pixel
                    dset[offset:offset+n,...] = data
            else:
                print(f"  Skipping {name} correction for shell {shell_nr} of {basename}")

    f.close()
    print("Shell %d done" % shell_nr)


def correct_maps_mpi(indir, yml_file, outdir, snapfile, basenames):

    if comm_rank == 0:

        # Get simulation directory absolute path
        sim_dir = os.path.abspath(os.path.dirname(yml_file))

        # Read simulation config file
        config = yaml.load(open(yml_file, "r"), yaml.loader.SafeLoader)
        
        # Get mean comoving neutrino density from the snapshot
        with h5py.File(snapfile, "r") as snap:
            omega_nu_0 = snap["Cosmology"].attrs["Omega_nu_0"][0]
            H0 = snap["Cosmology"].attrs["H0 [internal units]"][0]
            G = snap["PhysicalConstants"]["InternalUnits"].attrs["newton_G"][0]
        critical_density_0 = 3*(H0**2)/(8*np.pi*G)
        neutrino_density = omega_nu_0 * critical_density_0

        # Find paths to the shell redshift lists:
        # May be a different file for each lightcone
        radius_file = {}
        z_central = {}
        for basename in basenames:
            radius_file[basename] = None
            for i in range(8):
                section = "Lightcone%d" % i
                if section in config and config[section]["basename"] == basename and "radius_file" in config[section]:
                    radius_file[basename] = config[section]["radius_file"]
                    break
            if radius_file[basename] is None:
                radius_file[basename] = config["LightconeCommon"]["radius_file"]

            # Get absolute path to radius file
            if not os.path.isabs(radius_file[basename]):
                radius_file[basename] = os.path.normpath(os.path.join(sim_dir, radius_file[basename]))
            print("Taking shell redshifts for %s from: %s" % (basename, radius_file[basename]))

            # Read in the shell redshifts
            with open(radius_file[basename], "r") as f:
                header = f.readline()
                # Check for the expected header
                if header.strip() != "# Minimum redshift, Maximum redshift":
                    raise Exception("Shell radii must be specified as redshifts!")
                # Read radii
                z_central[basename] = []
                while True:
                    line = f.readline().strip()
                    if len(line) == 0:
                        break
                    else:
                        z_central[basename].append(0.5*sum([float(col) for col in line.split(",")]))

            z_central[basename] = np.asarray(z_central[basename], dtype=float)
            print("  Found %d shell redshifts" % len(z_central[basename]))
    
        # Accumulate correct_shell_file() arguments needed to process each shell
        args = []
        for basename in basenames:

            # Find number of shells in this lightcone
            with h5py.File("%s/%s_index.hdf5" % (indir, basename), "r") as infile:
                nr_shells = int(infile["Lightcone"].attrs["nr_shells"])
                nr_files_per_shell = int(infile["Lightcone"].attrs["nr_files_per_shell"])
            if nr_files_per_shell != 1:
                raise Exception("Can only correct single file outputs!")
            if nr_shells != len(z_central[basename]):
                raise Exception("Wrong number of shell redshifts!")

            # Ensure the output directory exists
            try:
                os.makedirs(outdir)
            except OSError:
                pass

            # Copy the index file to the output
            infile =  "%s/%s_index.hdf5" % (indir, basename)
            outfile = "%s/%s_index.hdf5" % (outdir, basename)
            shutil.copyfile(infile, outfile)

            # Store arguments to process this shell
            args += [(indir, basename, outdir, shell_nr, z_central[basename], neutrino_density)
                     for shell_nr in range(nr_shells)]

    # Make sure all output dirs have been created
    comm.barrier()

    # Apply corrections to all shells in parallel
    with MPICommExecutor(comm, root=0) as executor:
        if executor is not None:
            # Need to fetch results to ensure exception is raised on failure
            results = list(executor.starmap(correct_shell_file, args))


if __name__ == "__main__":
    
    usage="""
    Usage: python3 -m mpi4py lightcone_io_correct_maps.py \\
              basedir shell_redshifts outdir basename1 basename2 ...

    basedir      : location of the input lightcones
    shell_redshifts : name of file with the shell redshifts
    outdir       : where to write the output
    basename*    : basenames of lightcones to process
"""

    args = {}
    if comm_rank == 0:
        if len(sys.argv) < 5:
            print(usage)
            args = None
        else:
            args["indir"] = sys.argv[1]
            args["yml_file"] = sys.argv[2]
            args["outdir"] = sys.argv[3]
            args["snapfile"] = sys.argv[4]
            args["basenames"] = sys.argv[5:]
    args = comm.bcast(args)

    if args is None:
        MPI.Finalize()
        sys.exit(0)

    correct_maps_mpi(args["indir"], args["yml_file"], args["outdir"], args["snapfile"], args["basenames"])

    comm.barrier()
    if comm_rank == 0:
        print("All shells done for lightcones ", args["basenames"])



#!/bin/env python

import unyt

#
# Unit information which may be missing from lightcones
#
UNIT_CONV_NO_UNITS = {"U_"+base_unit : 0 for base_unit in "MLtIT"}
UNIT_CONV_PHOTON_FLUX_PER_UNIT_SURFACE = dict(UNIT_CONV_NO_UNITS, U_L=-2.0, U_t=-1.0)
UNIT_CONV_ENERGY_FLUX_PER_UNIT_SURFACE = dict(UNIT_CONV_NO_UNITS, U_M=1.0, U_t=-3.0)
missing_units = {
    "XrayErositaLowIntrinsicPhotons"   : UNIT_CONV_PHOTON_FLUX_PER_UNIT_SURFACE,
    "XrayErositaHighIntrinsicPhotons"  : UNIT_CONV_PHOTON_FLUX_PER_UNIT_SURFACE,
    "XrayROSATIntrinsicPhotons"        : UNIT_CONV_PHOTON_FLUX_PER_UNIT_SURFACE,
    "XrayErositaLowIntrinsicEnergies"  : UNIT_CONV_ENERGY_FLUX_PER_UNIT_SURFACE,
    "XrayErositaHighIntrinsicEnergies" : UNIT_CONV_ENERGY_FLUX_PER_UNIT_SURFACE,
    "XrayROSATIntrinsicEnergies"       : UNIT_CONV_ENERGY_FLUX_PER_UNIT_SURFACE,
}

# Name of attributes in InternalCodeUnits and Units groups
unit_cgs_name = {
    "U_I" : "Unit current in cgs (U_I)",
    "U_L" : "Unit length in cgs (U_L)",
    "U_M" : "Unit mass in cgs (U_M)",
    "U_T" : "Unit temperature in cgs (U_T)",
    "U_t" : "Unit time in cgs (U_t)",
}
cgs_unit = {
    "U_I" : "A",
    "U_L" : "cm",
    "U_M" : "g",
    "U_T" : "K",
    "U_t" : "s",
}

def read_cgs_units(infile):
    """
    Read CGS unit definitions from the Units group in a snapshot or lightcone
    """
    units_cgs = {}
    for name in unit_cgs_name:
        units_cgs[name] = infile["Units"].attrs[unit_cgs_name[name]][0]
    return units_cgs

def compute_cgs_factor(exponents, units_cgs):
    """
    Compute conversion to CGS using the Units group and the dimensions of a dataset
    """
    cgs_factor = 1.0
    for base_unit in units_cgs:
        exponent = exponents[base_unit]
        cgs_factor *= (units_cgs[base_unit]**exponent)
    return cgs_factor

def cgs_expression(exponents, units_cgs):

    if all([e==0.0 for e in exponents.values()]):
        return "[ - ]"

    expression = ""
    for base_unit, power in exponents.items():
        if power == 0:
            pass
        elif power == 1.0:
            expression += base_unit + " "
        elif power % 1.0 == 0.0:
            expression += ("%s^%d " % (base_unit, power))
        else:
            expression += ("%s^%7.4f " % (base_unit, power))

    expression += "[ "
    for base_unit, power in exponents.items():
        if power == 0:
            pass
        elif power == 1.0:
            expression += cgs_unit[base_unit] + " "
        elif power % 1.0 == 0.0:
            expression += ("%s^%d " % (cgs_unit[base_unit], power))
        else:
            expression += ("%s^%7.4f " % (cgs_unit[base_unit], power))
    expression += "]"

    return expression

def correct_units(dset, name, units_cgs):
    """
    Return corrected unit attributes for the specified dataset as a dict,
    which will be empty if there's nothing to do
    """

    if dset.attrs["a-scale exponent"] != 0 or dset.attrs["h-scale exponent"]:
        raise NotImplementedError("Only implemented for quantities with no a/h factors")

    corrections = {}
    if name in missing_units:

        # Find correct exponents
        for base_unit in units_cgs:
            expected_exponent = missing_units[name][base_unit]
            found_exponent = dset.attrs[base_unit+" exponent"][0]
            if found_exponent != expected_exponent:
                corrections[base_unit+" exponent"] = [expected_exponent,]

        # If any exponents changed, need to update CGS factor and expression
        if len(corrections) > 0:

            # Find new CGS conversion
            cgs_conversion = compute_cgs_factor(missing_units[name], units_cgs)
            corrections["Conversion factor to CGS (not including cosmological corrections)"] = [cgs_conversion,]
            corrections["Expression for physical CGS units"] = cgs_expression(missing_units[name], units_cgs)

    return corrections

def units_from_attributes(dset):
    cgs_factor = dset.attrs["Conversion factor to CGS (not including cosmological corrections)"][0]
    U_I = dset.attrs["U_I exponent"][0]
    U_L = dset.attrs["U_L exponent"][0]
    U_M = dset.attrs["U_M exponent"][0]
    U_T = dset.attrs["U_T exponent"][0]
    U_t = dset.attrs["U_t exponent"][0]
    return cgs_factor * (unyt.A**U_I) * (unyt.cm**U_L) * (unyt.g**U_M) * (unyt.K**U_T) * (unyt.s**U_t)



#!/bin/env python

import sys
import os
import shutil
import h5py
import numpy as np
import healpy as hp

from mpi4py import MPI
comm = MPI.COMM_WORLD
comm_rank = comm.Get_rank()
comm_size = comm.Get_size()

def shell_filename(basedir, basename, shell_nr, file_nr):
    return  ("%(basedir)s/%(basename)s_shells/shell_%(shell_nr)d/%(basename)s.shell_%(shell_nr)d.%(file_nr)d.hdf5" %
             {"basedir" : basedir, "basename" : basename,
              "shell_nr" : shell_nr, "file_nr" : file_nr})

def get_power(name):
    """
    Return the power parameter for hp.pixelfunc.ud_grade

    This determines if we compute the mean or sum of pixels
    when down-sampling.
    """
    SUM=-2
    MEAN=0
    power = {
        "TotalMass" : SUM,
        "DarkMatterMass" : SUM,
        "NeutrinoMass" : SUM,
        "SmoothedGasMass" : SUM,
        "UnsmoothedGasMass" : SUM,
        "StellarMass" : SUM,
        "BlackHoleMass" : SUM,
        "StarFormationRate" : SUM,
        "XrayErositaLowIntrinsicPhotons" : SUM,
        "XrayErositaLowIntrinsicEnergies" : SUM,
        "XrayErositaHighIntrinsicPhotons" : SUM,
        "XrayErositaHighIntrinsicEnergies" : SUM,
        "XrayROSATIntrinsicPhotons" : SUM,
        "XrayROSATIntrinsicEnergies" : SUM,
        "ComptonY" : MEAN,
        "DopplerB" : MEAN,
        "DM" : MEAN,
    }
    return power[name]

def downsample_maps(input_basedir, output_basedir, basename, shell_nr, new_nside):
    """
    This down samples all of the HEALPix maps in a single file.
    Only works on single file maps.
    """

    # Open the input file
    input_filename = shell_filename(input_basedir, basename, shell_nr, 0)
    print("Opening input: ", input_filename)
    infile = h5py.File(input_filename, "r")
    nr_files = infile["Shell"].attrs["nr_files_per_shell"]
    if nr_files != 1:
        raise Exception("Can only down sample single file outputs!")

    # Create the output file
    output_filename = shell_filename(output_basedir, basename, shell_nr, 0)
    print("Opening output: ", output_filename)
    os.makedirs(os.path.dirname(output_filename), exist_ok=True)
    outfile = h5py.File(output_filename, "w")

    # Copy metadata
    for group in ("InternalCodeUnits", "Units", "Shell"):
        infile.copy(group, outfile)

    # Loop over maps
    for name in infile:
        input_dataset = infile[name]
        if "nside" in input_dataset.attrs:
            # Read in the full map
            print("  Reading  %s for shell %d of %s" % (name, shell_nr, basename))
            nside = input_dataset.attrs["nside"][0]
            full_map_data = input_dataset[...]
            # Downsample the map
            print("  Sampling %s for shell %d of %s" % (name, shell_nr, basename))
            downsampled_map_data = hp.pixelfunc.ud_grade(full_map_data, new_nside, power=get_power(name))
            # Write the new map to the output file
            print("  Writing  %s for shell %d of %s" % (name, shell_nr, basename))
            output_dataset = outfile.create_dataset(name, data=downsampled_map_data, 
                                                    chunks=input_dataset.chunks,
                                                    compression=input_dataset.compression,
                                                    compression_opts=input_dataset.compression_opts)
            # Copy attributes
            for attr_name, attr_value in input_dataset.attrs.items():
                output_dataset.attrs[attr_name] = attr_value
            output_dataset.attrs["nside"] = (new_nside,)
            output_dataset.attrs["number_of_pixels"] = (len(downsampled_map_data),)

    print("Finished shell %d of %s" % (shell_nr, basename))
    outfile.close()
    infile.close()

if __name__ == "__main__":

    usage = """
    Usage:

    mpirun python3 -m mpi4py ./lightcone_io_downsample_maps.py \\
        input_dir output_dir basename new_nside
"""
    args = {}
    if comm_rank == 0:
        if len(sys.argv) == 5:
            args["input_basedir"]  = sys.argv[1]
            args["output_basedir"] = sys.argv[2]
            args["basename"]       = sys.argv[3]
            args["new_nside"]      = int(sys.argv[4])
        else:
            args = None
    args = comm.bcast(args)
    if args is None:
        if comm_rank == 0:
            print(usage)
        MPI.Finalize()
        sys.exit(0)

    # Determine number of shells
    if comm_rank == 0:
        input_index_filename = "%s/%s_index.hdf5" % (args["input_basedir"], args["basename"])
        with h5py.File(input_index_filename, "r") as infile:
            nr_shells = infile["Lightcone"].attrs["nr_shells"][0]
    else:
        nr_shells = None
    nr_shells = comm.bcast(nr_shells)

    # Parallelise over the shells
    for shell_nr in range(nr_shells):
        if shell_nr % comm_size == comm_rank:
            downsample_maps(args["input_basedir"], args["output_basedir"], args["basename"],
                            shell_nr, args["new_nside"])

    # Copy over the index file
    if comm_rank == 0:
        output_index_filename = "%s/%s_index.hdf5" % (args["output_basedir"], args["basename"])
        shutil.copyfile(input_index_filename, output_index_filename)

    comm.barrier()
    if comm_rank == 0:
        print("All shells done.")



#!/bin/env python

import socket
hostname = socket.gethostname()

try:
    import psutil
except ImportError:
    psutil = None


def report_usage(comm):
    """
    Report memory use of processes in the specified communicator
    """

    if psutil is None:
        return
    
    GB = 1024 ** 3

    # Get information about this node
    mem = psutil.virtual_memory()
    total_mem_gb = comm.allgather(mem.total / GB)
    free_mem_gb = comm.allgather(mem.available / GB)

    # Get information about this process
    process = psutil.Process()
    rss_gb  = comm.allgather(process.memory_info().rss / GB)

    # Report min and max
    if comm.Get_rank() == 0:
        min_free = min(free_mem_gb)
        max_free = max(free_mem_gb)
        min_rss  = min(rss_gb)
        max_rss  = max(rss_gb)
        print(f"Node free memory: min={min_free:.2f}GB, max={max_free:.2f}GB, process RSS: min={min_rss:.2f}GB, max={max_rss:.2f}GB")

        
if __name__ == "__main__":

    from mpi4py import MPI
    report_usage(MPI.COMM_WORLD)



#!/bin/env python

import sys

import h5py
import numpy as np
import lightcone_io.smoothed_map as smoothed_map
import virgo.mpi.parallel_hdf5 as phdf5

from mpi4py import MPI
comm = MPI.COMM_WORLD
comm_rank = comm.Get_rank()


def L1000N1800_HYDRO_FIDUCIAL_smoothed_gas_mass(shell_nr):

    nside = 16384
    input_filename = "/cosma8/data/dp004/jch/FLAMINGO/ScienceRuns/L1000N1800/HYDRO_FIDUCIAL/lightcones/lightcone0_particles/lightcone0_0000.0.hdf5"
    output_filename = f"/cosma8/data/dp004/jch/FLAMINGO/ScienceRuns/L1000N1800/HYDRO_FIDUCIAL/new_maps/smoothed_gas_map.shell_{shell_nr}.hdf5"
    shell_redshifts = "/cosma8/data/dp004/flamingo/Runs/L1000N1800/HYDRO_FIDUCIAL/shell_redshifts_z3.txt"

    # Read shell redshifts
    if comm_rank == 0:
        redshifts = np.loadtxt(shell_redshifts, delimiter=",")
        zmin = redshifts[shell_nr,0]
        zmax = redshifts[shell_nr,1]
        print(f"Reproducing shell {shell_nr} with zmin={zmin} and zmax={zmax}")
    else:
        zmin = None
        zmax = None
    zmin, zmax = comm.bcast((zmin, zmax))

    # Specify which particle types and properties to read in
    ptype = "Gas"
    property_names = ("Masses",)

    # Function to return the quantity to map, given the lightcone particle data dict
    def particle_mass(particle_data):
        return particle_data["Masses"]

    # Make the map
    map_data = smoothed_map.make_sky_map(input_filename, ptype, property_names, particle_mass,
                                         zmin, zmax, nside, smooth=True, progress=True)

    # Write out the new map
    dataset_name = "SmoothedGasMass"
    with h5py.File(output_filename, "w", driver="mpio", comm=comm) as outfile:
        phdf5.collective_write(outfile, dataset_name, map_data, comm)



if __name__ == "__main__":

    # Shell to do should be specified on the command line
    if comm_rank == 0:
        shell_nr = int(sys.argv[1])
    else:
        shell_nr = None
    shell_nr = comm.bcast(shell_nr)

    L1000N1800_HYDRO_FIDUCIAL_smoothed_gas_mass(shell_nr)



#!/bin/env python
#
# This script plots a HEALPix map
#

import matplotlib.pyplot as plt
import healpy as hp
import numpy as np

import lightcone_io.healpix_maps as hm

# Lightcone base directory
basedir = "/cosma8/data/dp004/jch/FLAMINGO/ScienceRuns/L1000N1800/DMO_FIDUCIAL/lightcones/"

# Lightcone base name
basename = "lightcone0"

# Which shell to plot
shell_nr = 10

# Which map to plot
map_name = "TotalMass"

# Open the lightcone map set
shell = hm.ShellArray(basedir, basename)

print("Comoving inner radius = ", shell[shell_nr].comoving_inner_radius)
print("Comoving outer radius = ", shell[shell_nr].comoving_outer_radius)

# Read the map
map_data = shell[shell_nr][map_name][...]

# Plot the map
hp.mollview(map_data, norm="log")
plt.show()
